// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.gui;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import de.uka.ilkd.key.proof.SLListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.IteratorOfNode;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Node;
import org.apache.log4j.Logger;



public class KeYSelectionModel {

    /** the proof to listen to */
    private Proof proof;
    /** if true the selectedGoal field below can be trusted */
    private boolean goalIsValid;
    /** is the selected node a goal */
    private Goal selectedGoal;
    /** the current displayed node */
    private Node selectedNode;
    /** the listeners to this model */
    private List listenerList;
    /** cached selected node event */
    private KeYSelectionEvent selectionEvent = 
	new KeYSelectionEvent(this);
    
    private Logger threadLogger = Logger.getLogger("key.threading");

    
    KeYSelectionModel() {
	listenerList = Collections.synchronizedList(new ArrayList(5));
	goalIsValid = false;
    }

    /** sets the selected proof 
     * @param p the Proof that is selected
     */
    public void setSelectedProof(Proof p) {
	goalIsValid = false;
	proof = p;
        if (proof !=null) {
	    Goal g = proof.openGoals().iterator().next();	    
	    if ( g == null) {
	        selectedNode = proof.root().leavesIterator().next();
	    } else {
	        goalIsValid = true;
	        selectedNode = g.node();
	        selectedGoal = g;
	    }
        } else {
            selectedNode = null;
            selectedGoal = null;
        }
	fireSelectedProofChanged();
    }

    /** returns the proof that is selected by the user
     * @return  the proof that is selected by the user
     */ 
    public Proof getSelectedProof() {
	return proof;
    }

    /** sets the node that is focused by the user
     * @param n the selected node
     */
    public void setSelectedNode(Node n) {
	goalIsValid = false;
	selectedNode = n;
	fireSelectedNodeChanged();
    }

    /** sets the node that is focused by the user
     * @param g the Goal that contains the selected node
     */
    public void setSelectedGoal(Goal g) {
	goalIsValid = true;
	selectedGoal = g;
	selectedNode = g.node();
	fireSelectedNodeChanged();
    }

    /** returns the node that is selected by the user
     * @return  the node that is selected by the user
     */ 
    public Node getSelectedNode() {
	return selectedNode;
    }

    /** returns the goal the selected node belongs to, null if it is
     * an inner node
     * @return  the goal the selected node belongs to, null if it is
     * an inner node
     */ 
    public Goal getSelectedGoal() {
	if (proof==null) {
	    throw new IllegalStateException("No proof loaded.");
	}
	if (!goalIsValid) {
	    selectedGoal = proof.getGoal(selectedNode);
	    goalIsValid = true;
	} 
	return selectedGoal;
    }

    /** returns true iff the selected node is a goal
     * @return true iff the selected node is a goal
     */
    public boolean isGoal() {
	if (!goalIsValid) {
	    return (getSelectedGoal() != null);
	}	
	return selectedGoal != null;
    }

   
    /** enumerate the possible goal selections, starting with the best
     * one
     */
    protected class DefaultSelectionIterator implements IteratorOfGoal {
	private static final int POS_START     = 0;
	private static final int POS_LEAVES    = 1;
	private static final int POS_GOAL_LIST = 2;

	private int            currentPos = POS_START;
	private Goal           nextOne;
	private IteratorOfGoal goalIt;
	private IteratorOfNode nodeIt;

	public  DefaultSelectionIterator () {
	    findNext ();
	}

	private void findNext            () {
	    nextOne = null;
	    while ( nextOne == null ) {
		switch ( currentPos ) {
		case POS_START:
		    currentPos = POS_LEAVES;
		    if ( selectedNode != null )
			nodeIt = selectedNode.leavesIterator ();
		    else
			nodeIt = null;
		    break;
		case POS_LEAVES:
		    if ( nodeIt == null || !nodeIt.hasNext () ) {
			currentPos = POS_GOAL_LIST;
			if ( proof.openGoals() != SLListOfGoal.EMPTY_LIST )
			    goalIt = proof.openGoals().iterator();
			else
			    goalIt = null;
		    } else
			nextOne = proof.getGoal ( nodeIt.next() );
		    break;

		case POS_GOAL_LIST:
		    if ( goalIt == null || !goalIt.hasNext () )
			// no more items
			return;
		    else
			nextOne = goalIt.next();
		    break;
		}
	    }
	}

	public  Goal    next             () {
	    Goal res = nextOne;
	    findNext ();
	    return res;
	}

	public  boolean hasNext          () {
	    return nextOne != null;
	}
    }


    /** selectes the first goal in the goal list of proof if available
     * if not it selectes a leaf of the proof tree
     */
    public void defaultSelection() {		
	Goal           g       = null;
	Goal           firstG  = null;
	IteratorOfGoal it      = new DefaultSelectionIterator ();
	
	while ( g == null && it.hasNext () ) {
	    g = it.next ();
	    if ( firstG == null )
		firstG = g;
	    if ( g.getClosureConstraint ().isSatisfiable () )
		g = null;
	}

	/** Order of preference:
	 * 1. Not yet closable goals
	 * 2. Goals which are not closed for all metavariable
	 * instantiations
	 * 3. The first node of the tree
	 */
	if ( g != null )
	    setSelectedGoal(g);
	else {
	    if ( firstG != null )
		setSelectedGoal(firstG);
	    else
		setSelectedNode(proof.root().leavesIterator().next());
	}
	/*
	if (selectedNode != null) {
	    IteratorOfNode nodeIt = selectedNode.leavesIterator();	
	    while (nodeIt.hasNext()) {
		g = proof.getGoal(nodeIt.next());
		if (g != null) {
		    break;
		}
	    }
	}
	if (g == null && proof.openGoals() != SLListOfGoal.EMPTY_LIST ) {	       	    
	    g = proof.openGoals().iterator().next();
	}
	if (g != null) {
	    setSelectedGoal(g);
	} else {
	    setSelectedNode(proof.root().leavesIterator().next());	
	}
	*/
    }

    public void addKeYSelectionListener(KeYSelectionListener listener) {
	synchronized(listenerList) {
	    threadLogger.info("Adding "+listener.getClass());
	    listenerList.add(listener);	    
	}
    }

    public void removeKeYSelectionListener(KeYSelectionListener listener) {
	synchronized(listenerList) {
	    threadLogger.info("Removing "+listener.getClass());
	    listenerList.remove(listener);	    
	}
    }		

    public synchronized void fireSelectedNodeChanged() {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((KeYSelectionListener)it.next()).
		    selectedNodeChanged(selectionEvent);
	    }
	}
    }

    public synchronized void fireSelectedProofChanged() {
	synchronized(listenerList) {
	    threadLogger.info("Selected Proof changed, firing...");
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((KeYSelectionListener)it.next()).
		    selectedProofChanged(selectionEvent);
	    }
	    threadLogger.info("Selected Proof changed, done firing.");
	}
    }    
    
}
