// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.gui;

import javax.swing.*;

import java.awt.event.*;
import java.util.*;
import java.io.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.pp.*;
import de.uka.ilkd.key.util.pp.*;
import de.uka.ilkd.key.rule.*;
import org.apache.log4j.Logger;

/**
 *  This class creates a menu with Taclets as entries. The invoker has
 * to be of type SequentView because of the method call selectedTaclet
 * that hands over the selected Taclet. The class is used to get all
 * Taclet that are applicable at a selected position in a sequent.
 */ 
class TacletMenu extends JMenu {

    private PosInSequent pos;
    private SequentView sequentView;
    private KeYMediator mediator;

    private TacletAppComparator comp = new TacletAppComparator();
    
    static Logger logger = Logger.getLogger(TacletMenu.class.getName());
    
    /** creates empty menu 
     */
    TacletMenu() {
    }


    /** creates a new menu that displays all applicable rules at the given
     * position
     * @param sequentView the SequentView that is the parent of this menu
     * @param findList ListOfTaclet with all applicable FindTaclets
     * @param rewriteList ListOfTaclet with all applicable RewriteTaclets
     * @param noFindList ListOfTaclet with all applicable noFindTaclets
     * @param builtInList ListOfBuiltInRule with all applicable BuiltInRules
     * @param pos the PosInSequent
     */ 
    TacletMenu(SequentView sequentView,
	       ListOfTacletApp findList, ListOfTacletApp rewriteList,
	       ListOfTacletApp noFindList, ListOfBuiltInRule builtInList,
	       PosInSequent pos) {
        super();        
	this.sequentView = sequentView;
	this.mediator = sequentView.mediator();
 	this.pos = pos;
	// delete RewriteTaclet from findList because they will be in
	// the rewrite list and concatenate both lists
	createTacletMenu(removeRewrites(findList).prepend(rewriteList),
			 noFindList, builtInList, new MenuControl());
    }

    
    /** removes RewriteTaclet from list
     * @param list the ListOfTaclet from where the RewriteTaclet are
     * removed
     * @return list without RewriteTaclets
     */
    private ListOfTacletApp removeRewrites(ListOfTacletApp list) {
	ListOfTacletApp result = SLListOfTacletApp.EMPTY_LIST;
	IteratorOfTacletApp it = list.iterator();

	while(it.hasNext()) {
	    TacletApp tacletApp = it.next();
	    Taclet taclet=tacletApp.taclet();
	    result = (taclet instanceof RewriteTaclet ? result :
		      result.prepend(tacletApp));
	}
	return result;
    }


    /** creates the menu by adding all submenus and items */
    private void createTacletMenu(ListOfTacletApp find,
				  ListOfTacletApp noFind,
				  ListOfBuiltInRule builtInList,
				  MenuControl control) {	 
	addActionListener(control);
	boolean rulesAvailable=(addSection("Find", sort(find), control));
	if (pos != null && pos.isSequent()) {
	    rulesAvailable=addSection("NoFind", noFind, control)
		| rulesAvailable;
	}
	if (!rulesAvailable) {
	    createSection("No rules applicable.");
	}

	createBuiltInRuleMenu(builtInList, control);
        
	createFocussedAutoModeMenu ( control );
    
	//        addPopFrameItem(control);

	addClipboardItem(control);

	if (pos != null) {
	    PosInOccurrence occ = pos.getPosInOccurrence();	    
	    if (occ != null && occ.posInTerm() != null) {
		Term t = occ.subTerm ();
		createAbbrevSection(t, control);

		if(t.op() instanceof ProgramVariable) {
		    ProgramVariable var = (ProgramVariable)(t.op());
		    if(var.getProgramElementName().getCreationInfo() != null) {
		    	createNameCreationInfoSection(control);
		    }
		}
	    }
	}
    }

    private void createBuiltInRuleMenu(ListOfBuiltInRule builtInList,
				       MenuControl            control) {

	if (builtInList != SLListOfBuiltInRule.EMPTY_LIST) {
	    addSeparator();	
	    IteratorOfBuiltInRule it = builtInList.iterator();
	    while (it.hasNext()) {
		addBuiltInRuleItem(it.next(), control);
	    }
	}
    }
				      
    /**
     * adds an item for built in rules (e.g. Run Simplify or Update Simplifier)
     */
    private void addBuiltInRuleItem(BuiltInRule builtInRule,
				    MenuControl control) {
        JMenuItem item = new BuiltInRuleMenuItem(builtInRule);
        item.addActionListener(control);
	add(item);
    }

    
    private void createFocussedAutoModeMenu (MenuControl control) {
        addSeparator();
        JMenuItem item = new FocussedRuleApplicationMenuItem ();
        item.addActionListener(control);
        add(item);        
    }
    

    private ListOfTacletApp sort(ListOfTacletApp finds) {
	ListOfTacletApp result = SLListOfTacletApp.EMPTY_LIST;
	List list = new ArrayList(finds.size());

	IteratorOfTacletApp it = finds.iterator();
	while (it.hasNext()) {
	    list.add(it.next());
	}

	Collections.sort(list, comp);

	Iterator it2 = list.iterator();
	while (it2.hasNext()) {
	    result = result.prepend((TacletApp)it2.next());
	}

	return result;
    }


    private void createAbbrevSection(Term t, MenuControl control){
	AbbrevMap scm = mediator.getNotationInfo().getAbbrevMap();
	JMenuItem sc = null;
	if(scm.containsTerm(t)){
	    sc = new JMenuItem("Change abbreviation");
	    sc.addActionListener(control);
	    add(sc);
	    if(scm.isEnabled(t)){
		sc = new JMenuItem("Disable abbreviation");
	    }else{
		sc = new JMenuItem("Enable abbreviation");
	    }
	}else{
	    sc = new JMenuItem("Create abbreviation");
	}
	sc.addActionListener(control);
	add(sc);
    }


    private void createNameCreationInfoSection(MenuControl control) {
	JMenuItem item = new JMenuItem("View name creation info...");
	item.addActionListener(control);
	add(item);
    }



    /** creates a non selectable label with the specified name and adds it to
     * the component followed by the entries of this section 
     * @param title a String the title of the section
     * @param taclet ListOfTaclet that contains the Taclets belonging to this section
     * @return true if section has been added (empty sections are not added)
     */ 
    private boolean addSection(String title, ListOfTacletApp taclet, 
			       MenuControl control) {
	if (taclet.size() > 0) {
	    //uncomment if you want submenus with subtitels
	    //	    insert(createSubMenu(taclet, title, control), 1);
	    //	    createSection(title);
	    add(createMenuItems(taclet, control));
	    return true;
	}
	return false;
    }

    /** inserts separator followed from the section's title 
     * @param title a String that contains the title of the section
     */
    private void createSection(String title) {
	//addSeparator();
	add(new JLabel(title));
    }
    

    private void addPopFrameItem(MenuControl control) {
	JMenuItem item = new JMenuItem("Pop method frame");
	item.addActionListener(control);
	add(item);
    }

    
    private void addClipboardItem(MenuControl control) {
	addSeparator();
	JMenuItem item = new JMenuItem("to clipboard");
	item.addActionListener(control);
	add(item);
    }
    


    /** adds array of TacletMenuItem to itself*/
    private void add(TacletMenuItem[] items) {
	for (int i = 0; i < items.length; i++) {
	    add(items[i]);
	}
    }

    /** creates new TacletMenuItems for each taclet in the list and set
     * the given MenuControl as their ActionListener
     * @param taclet ListOfTaclet with the Taclets the items represent
     * @param control the ActionListener
     * @return the new MenuItems
     */
    private TacletMenuItem[] createMenuItems(ListOfTacletApp taclet, 
					     MenuControl  control) {
	TacletMenuItem[] items = new TacletMenuItem[taclet.size()];
	IteratorOfTacletApp it = taclet.iterator();
	
	for (int i = 0; it.hasNext(); i++) {
	    items[i] = new TacletMenuItem(it.next(), mediator.getNotationInfo());
	    items[i].addActionListener(control);
	}
	return items;
    }
        
    /** makes submenus invisible */
    void invisible() {
	for (int i = 0; i < getMenuComponentCount(); i++) {
	    if (getMenuComponent(i) instanceof JMenu) 
		((JMenu)getMenuComponent(i)).getPopupMenu().setVisible(false);
	}
    }
    
    /** ActionListener */
    class MenuControl implements ActionListener{

	private boolean validabbreviation(String s){
	    if(s==null || s.length()==0) return false;
	    for(int i=0; i<s.length(); i++){
		if(!((s.charAt(i)<='9' && s.charAt(i)>='0')||
		     (s.charAt(i)<='z' && s.charAt(i)>='a')||
		     (s.charAt(i)<='Z' && s.charAt(i)>='A')||
		     s.charAt(i)=='_')) return false;
	    }
	    return true;
	}

	public void actionPerformed(ActionEvent e) {
	    if (e.getSource() instanceof TacletMenuItem) {
		((SequentView)(getPopupMenu().getInvoker()))
		    .selectedTaclet(((TacletMenuItem) e.getSource()).connectedTo(), 
				    pos);
	    } else if (e.getSource() instanceof BuiltInRuleMenuItem) {
		mediator.selectedBuiltInRule
		    (((BuiltInRuleMenuItem) e.getSource()).connectedTo(), 
		     pos.getPosInOccurrence());
	    } else if (e.getSource() instanceof FocussedRuleApplicationMenuItem) {
	        mediator.getInteractiveProver ()
	            .startFocussedAutoMode ( pos.getPosInOccurrence (),
	                                     mediator.getSelectedGoal () );
	    } else {
		if (((JMenuItem)e.getSource()).getText()
		    .startsWith("to clipboard")){
                    Main.copyHighlightToClipboard(sequentView);
		} else if(((JMenuItem)e.getSource()).getText().
			  startsWith("Pop method frame")){
		    //                        mediator.popMethodFrame();
		} else if(((JMenuItem)e.getSource()).getText().
			  startsWith("Disable abbreviation")){
		    PosInOccurrence occ = pos.getPosInOccurrence();	    
		    if (occ != null && occ.posInTerm() != null) {
			mediator.getNotationInfo().getAbbrevMap().setEnabled(occ.subTerm(),false);
			sequentView.printSequent();
		    }
		}else if(((JMenuItem)e.getSource()).getText().
			 startsWith("Enable abbreviation")){
		    PosInOccurrence occ = pos.getPosInOccurrence();	    
		    if (occ != null && occ.posInTerm() != null) {
			mediator.getNotationInfo().
			    getAbbrevMap().setEnabled(occ.subTerm(),true);
			sequentView.printSequent();
		    }
		}else if(((JMenuItem)e.getSource()).getText().
			 startsWith("Create abbreviation")){
		    PosInOccurrence occ = pos.getPosInOccurrence();
		    if (occ != null && occ.posInTerm() != null) {
			String abbreviation = (String)JOptionPane.showInputDialog
			    (new JFrame(),
			     "Enter abbreviation for term: \n"+occ.subTerm().toString(), 
			     "New Abbreviation",
			     JOptionPane.QUESTION_MESSAGE,
			     null,
			     null,
			     "");
				    
			try{
			    if(abbreviation!=null){
				if(!validabbreviation(abbreviation)){
				    JOptionPane.showMessageDialog(new JFrame(),
								  "Only letters, numbers and '_' are allowed for Abbreviations", 
								  "Sorry",
								  JOptionPane.INFORMATION_MESSAGE);
				}else{
				    mediator.getNotationInfo().
					getAbbrevMap().put(occ.subTerm(),abbreviation,true);
				    sequentView.printSequent();
				}
			    }
			}catch(AbbrevException sce){
			    JOptionPane.showMessageDialog(new JFrame(), sce.getMessage(), "Sorry",
							  JOptionPane.INFORMATION_MESSAGE);
			}
		    }

		}else if(((JMenuItem)e.getSource()).getText().
			 startsWith("Change abbreviation")){
		    PosInOccurrence occ = pos.getPosInOccurrence();
		    if (occ != null && occ.posInTerm() != null) {
			String abbreviation = (String)JOptionPane.showInputDialog
			    (new JFrame(),
			     "Enter abbreviation for term: \n"+occ.subTerm().toString(),
			     "Change Abbreviation",
			     JOptionPane.QUESTION_MESSAGE,
			     null,
			     null,
			     mediator.getNotationInfo().
			     getAbbrevMap().getAbbrev(occ.subTerm()).substring(1));
			try{
			    if(abbreviation!=null){
				if(!validabbreviation(abbreviation)){
				    JOptionPane.showMessageDialog(new JFrame(),
								  "Only letters, numbers and '_' are allowed for Abbreviations",
								  "Sorry",
								  JOptionPane.INFORMATION_MESSAGE);
				}else{
				    mediator.getNotationInfo().
					getAbbrevMap().changeAbbrev(occ.subTerm(),abbreviation);
				    sequentView.printSequent();
				}
			    }
			}catch(AbbrevException sce){
			    JOptionPane.showMessageDialog(new JFrame(), sce.getMessage(), "Sorry",
							  JOptionPane.INFORMATION_MESSAGE);
			}
		    }
		} else if(((JMenuItem)e.getSource()).getText().
			 startsWith("View name creation info")) {
		    Term t = pos.getPosInOccurrence().subTerm();
		    ProgramVariable var = (ProgramVariable)t.op();
		    ProgramElementName name = var.getProgramElementName();
		    NameCreationInfo info = name.getCreationInfo();
		    String message;
		    if(info != null) {
			message = info.infoAsString();
		    } else {
		        message = "No information available.";
		    }
		    JOptionPane.showMessageDialog(null,
		    				  message,
						  "Name creation info",
		  				  JOptionPane.INFORMATION_MESSAGE);
		}
	    }
	}
    }



    /** 
     * this class extends JMenuItem. The objective is to store
     * the Taclet of each item in the item for easier access to the Taclet
     * if the item has been selected 
     */
    static class TacletMenuItem extends JMenuItem {
	
	private TacletApp connectedTo;
	
	/** creates TacletMenuItem attached to a Taclet 
	 * @param connectedTo the TacletApp that is represented by the item 
	 * @param notationInfo the NotationInfo used to print terms
	 */
	public TacletMenuItem(TacletApp connectedTo, NotationInfo notationInfo) {
	    super(connectedTo.taclet().displayName());
	    this.connectedTo = connectedTo;	    	    
	    StringBuffer taclet_sb = new StringBuffer();
	    Writer w = new StringWriter();
	    
	    WriterBackend backend = new WriterBackend(w, 68);
	    LogicPrinter tp = new LogicPrinter(new ProgramPrinter(w,
								  connectedTo.instantiations()),
					       notationInfo, backend,
					       true);
	    tp.printTaclet(connectedTo.taclet(), connectedTo.instantiations(),
			   ProofSettings.DEFAULT_SETTINGS.getViewSettings()
			   .getShowWholeTaclet());
	    
	    int nlcount = 0;
	    StringBuffer sb = new StringBuffer(w.toString());
	    int sbl = sb.length();
	    for (int i = 0; i < sbl; i++) {
		if (sb.charAt(i) == '\n') {
		    nlcount += 1;
		}
	    }
	    if (nlcount > ProofSettings.DEFAULT_SETTINGS.getViewSettings().getMaxTooltipLines()) { 
		if (logger.isDebugEnabled()) {
		    logger.debug("No SchemaVariable instantiation printed, linecount is " 
				 + nlcount + " > " 
				 + ProofSettings.DEFAULT_SETTINGS.
				 getViewSettings().getMaxTooltipLines());
		}
		taclet_sb = new StringBuffer();
		w = new StringWriter();
		backend = new WriterBackend(w, 68);
		tp = new LogicPrinter
		    (new ProgramPrinter(w), notationInfo, backend, true);
		tp.printTaclet(connectedTo.taclet());

	    }

	    taclet_sb.append(ascii2html(new StringBuffer(w.toString())));
	    taclet_sb.replace(0,0,"<html><pre>");
	    taclet_sb.append("</pre>");


	    setToolTipText(taclet_sb.toString());
	} 
	
	/**
	 * Replaces <,>,& and new lines with their HTML masks.
	 * @param sb The StringBuffer with forbidden HTML characters
	 * @return A new StringBuffer with the masked characters.
	 */
	protected StringBuffer ascii2html(StringBuffer sb) {
	    StringBuffer nsb = new StringBuffer();
	    StringBuffer asb = removeEmptyLines(sb);
	    int sbl = asb.length();
	    for (int i = 0; i < sbl; i++) {
	    	switch (asb.charAt(i)) {
		case '<'	: nsb.append("&lt;"); break;
		case '>'	: nsb.append("&gt;"); break;
		case '&'	: nsb.append("&amp;"); break;
		case '\n'	: nsb.append("<br>"); break;
		default		: nsb.append(asb.charAt(i));
	    	}
	    }
	    return nsb;
	}

	protected StringBuffer removeEmptyLines(StringBuffer sb) {
	    String s = sb.toString();
	    String[] sa = s.split("\n");
	    StringBuffer result = new StringBuffer();
	    for (int i = 0; i < sa.length; i ++) {
		//logger.debug("'" + sa[i] + "'");
		if (sa[i] == "") {
		    continue;
		}
		boolean onlySpaces = true;
		for (int j = 0; j < sa[i].length(); j++) {
		    if (sa[i].charAt(j) != ' ') {
			onlySpaces = false;
		    }
		}
		if (onlySpaces) {
		    continue;
		}
		result.append(sa[i]).append("\n");
	    }
	    if (result.charAt(result.length()-1) == '\n') {
		result.setLength(result.length() - 1);
	    }
	    return result;
	}
	
	/** gets the Taclet that is attached to this item
	 * @return the attached Taclet
	 */
	public TacletApp connectedTo() {
	    return connectedTo;
	}

    }

    /** 
     * equal to TacletMenuItem but for BuiltInRules
     */
    static class BuiltInRuleMenuItem extends JMenuItem {
	
	private BuiltInRule connectedTo;
	
	public BuiltInRuleMenuItem(BuiltInRule connectedTo) {
	    super(connectedTo.name().toString());
	    this.connectedTo = connectedTo;	    	    
	} 

	public BuiltInRule connectedTo() {
	    return connectedTo;
	}

    }
    
    static class FocussedRuleApplicationMenuItem extends JMenuItem {
        public FocussedRuleApplicationMenuItem () {
            super("Apply rules automatically here ...");
        }
    }
    
    
    static class TacletAppComparator implements Comparator {

	private int countFormulaSV(TacletSchemaVariableCollector c) {
	    int formulaSV = 0;
	    IteratorOfSchemaVariable it = c.varIterator();
	    while (it.hasNext()) {
		SchemaVariable sv = it.next();
		if(sv instanceof SortedSchemaVariable) {
		    if (((SortedSchemaVariable)sv).sort() == Sort.FORMULA) {
			formulaSV++;
		    }
		}
	    }
	
	    return formulaSV;
	}
	
	/** this is a rough estimation about the goal complexity. The
	 * complexity depends on the depth of the term to be replaced.
	 * If no such term exists we add a constant (may be refined in
	 * future)
	 */
	private int measureGoalComplexity(ListOfTacletGoalTemplate l) {
	    int result = 0;
	    IteratorOfTacletGoalTemplate it = l.iterator();
	    while (it.hasNext()) {
		TacletGoalTemplate gt = it.next();
		if (gt instanceof RewriteTacletGoalTemplate) {
		    if (((RewriteTacletGoalTemplate)gt).replaceWith() != null) {
			result += ((RewriteTacletGoalTemplate)gt).replaceWith().depth();
		    }
		} 
		if (gt.sequent() != Sequent.EMPTY_SEQUENT) {
		    result += 10;
		}
	    }
	    return result;
	}
	
	
	/**
	 * rough approximation of the program complexity
	 */
	public int programComplexity(JavaBlock b) {
	    if (b == JavaBlock.EMPTY_JAVABLOCK) {
		return 0;
	    }
	    return new de.uka.ilkd.key.java.visitor.JavaASTWalker(b.program()) { 
		    private int counter = 0;
			    
		    protected void doAction(ProgramElement pe) {
			counter++;
		    }
			    
		    public int getCounter() {
			counter = 0;
			start();
			return counter;
		    }
		}.getCounter();
	}
	
	public int compare(Object o1, Object o2) {
	    FindTaclet taclet1 = (FindTaclet)(((TacletApp)o1).taclet());
	    FindTaclet taclet2 = (FindTaclet)(((TacletApp)o2).taclet());
		    
	    int findComplexity1 = taclet1.find().depth();
	    int findComplexity2 = taclet2.find().depth();
	    findComplexity1 += programComplexity(taclet1.find().javaBlock());
	    findComplexity2 += programComplexity(taclet2.find().javaBlock());
	
	    if ( findComplexity1 < findComplexity2 ) {
		return -1;
	    } else if (findComplexity1 > findComplexity2) {
		return 1;
	    }
		    
		    
		    
	
	    // depth are equal. Number of schemavariables decides
	    TacletSchemaVariableCollector coll1 = new TacletSchemaVariableCollector();
	    taclet1.find().execPostOrder(coll1);
	    int formulaSV1 = countFormulaSV(coll1);
		    
	    TacletSchemaVariableCollector coll2  = new TacletSchemaVariableCollector();
	    taclet2.find().execPostOrder(coll2);
	    int formulaSV2 = countFormulaSV(coll2);
	
	    int cmpVar1 = -coll1.size()+taclet1.getRuleSets().size();
	    int cmpVar2 = -coll2.size()+taclet2.getRuleSets().size();
		    
	    if (cmpVar1 == cmpVar2) {
		cmpVar1 = cmpVar1-formulaSV1;
		cmpVar2 = cmpVar2-formulaSV2;
	    }
		    
	    if ( cmpVar1 < cmpVar2 ) {
		return -1;
	    } else if (cmpVar1 > cmpVar2) {
		return 1;
	    }
	
	    if (taclet1.ifSequent() == Sequent.EMPTY_SEQUENT && 
		taclet2.ifSequent() != Sequent.EMPTY_SEQUENT) {
		return 1;
	    } else if (taclet1.ifSequent() == Sequent.EMPTY_SEQUENT && 
		       taclet2.ifSequent() != Sequent.EMPTY_SEQUENT) {
		return -1;
	    }
		    
	    int goals1 = -taclet1.goalTemplates().size();
	    int goals2 = -taclet2.goalTemplates().size();
	
	    if (goals1 == goals2) {
		goals1 = -measureGoalComplexity(taclet1.goalTemplates());
		goals2 = -measureGoalComplexity(taclet2.goalTemplates());		
	    } 
	
	    if (goals1 < goals2) {
		return -1;
	    } else if (goals1 > goals2) {
		return 1;
	    }
	
	    return 0;
	}
	
    }
    
}
