// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.java;

import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.java.recoderext.SchemaCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.logic.NamespaceSet;
import de.uka.ilkd.key.logic.VariableNamer;
import de.uka.ilkd.key.logic.InnerVariableNamer;
import de.uka.ilkd.key.logic.OuterVariableNamer;
import de.uka.ilkd.key.jml.Implementation2SpecMap;
import de.uka.ilkd.key.proof.Counter;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.KeYExceptionHandler;
import de.uka.ilkd.key.util.KeYRecoderExcHandler;
import java.util.HashMap;
import java.util.Iterator;

/**
 * this is a collection of common services to the KeY prover. Services
 * include information on the underlying Java model and a converter to
 * transform Java program elements to logic (where possible) and back.
 */
public class Services{

    /**
     * proof specific namespaces (functions, predicates, sorts, variables)
     */
    private NamespaceSet namespaces = new NamespaceSet();

    /** used to determine whether an expression is a compile-time 
     * constant and if so the type and result of the expression
     */
    private ConstantExpressionEvaluator cee;

    /** used to convert types, expressions and so on to logic elements
     * (in special into to terms or formulas)
     */
    private TypeConverter typeconverter;

    /**
     * the information object on the Java model
     */
    private JavaInfo javainfo;

    /**
     * mapping for jml specifications
     */
    private Implementation2SpecMap specMap;

    /**
     * variable namer for inner renaming
     */
    private VariableNamer innerVarNamer = new InnerVariableNamer(this);

    /**
     * variable namer for outer renaming
     */
    private VariableNamer outerVarNamer = new OuterVariableNamer(this);

    /**
     * current status of renaming strategy selection
     */
    private static boolean outerRenaming = false;

    /**
     * the exception-handler
     */
    private KeYExceptionHandler exceptionHandler;
    

    /**
     * map of names to counters
     */
    private HashMap counters = new HashMap();


    /**
     * creates a new Services object with a new TypeConverter and a new
     * JavaInfo object with no information stored at none of these.
     */
    public Services(KeYExceptionHandler exceptionHandler){
	cee = new ConstantExpressionEvaluator(this);
        typeconverter = new TypeConverter(this);
	if(exceptionHandler == null){
	    this.exceptionHandler = new KeYRecoderExcHandler();
	}else{
	    this.exceptionHandler = exceptionHandler;
	}
        javainfo = new JavaInfo
        (new KeYProgModelInfo(typeconverter, exceptionHandler), this);
	specMap = new Implementation2SpecMap(this);
    }

    public Services(){
	this((KeYExceptionHandler) null);
    }

    private Services(KeYCrossReferenceServiceConfiguration crsc, 
		     KeYRecoderMapping rec2key) {
	cee = new ConstantExpressionEvaluator(this);
	typeconverter = new TypeConverter(this);
	//	exceptionHandler = new KeYRecoderExcHandler();
	javainfo = new JavaInfo
	    (new KeYProgModelInfo(crsc, rec2key, typeconverter), this);
    }


     /** THIS CONSTRUCTOR IS ONLY FOR TESTS!
      * creates a Services object that contains the given JavaInfo object, a
      * type converter is created
      * @param ji the JavaInfo to use
      */
     public Services(JavaInfo ji){
         javainfo = ji;
         typeconverter = new TypeConverter(this);
	 specMap = new Implementation2SpecMap(this);
	 exceptionHandler = new KeYRecoderExcHandler();
     }

    public KeYExceptionHandler getExceptionHandler(){
	return exceptionHandler;
    }

    public void setExceptionHandler(KeYExceptionHandler keh){
	exceptionHandler = keh;
    }

    /**
     * Returns the TypeConverter associated with this Services object.
     */
    public TypeConverter getTypeConverter(){
        return typeconverter;
    }

    private void setTypeConverter(TypeConverter tc) {
	typeconverter=tc;
    }

    /**
     * Returns the ConstantExpressionEvaluator associated with this Services object.
     */
    public ConstantExpressionEvaluator getConstantExpressionEvaluator(){
        return cee;
    }

    private void setConstantExpressionEvaluator(ConstantExpressionEvaluator ce) {
	cee = ce;
    }

    /**
     * Returns the JavaInfo associated with this Services object.
     */
    public JavaInfo getJavaInfo(){
        return javainfo;
    }

    private void setJavaInfo(JavaInfo ji) {
	javainfo=ji;
	if (ji.getServices()!=this) {
	    throw new IllegalStateException("Services are inconsistent.");
	}
    }


    /**
     * Returns the VariableNamer associated with this Services object.
     */
    public VariableNamer getVariableNamer() {
        return (outerRenaming ? outerVarNamer : innerVarNamer);
    }
    
    public static void setOuterRenaming(boolean enabled) {
    	outerRenaming = enabled;
    }


    /**
     * Returns the Implementation2SpecMap associated with this Services object.
     */
    public Implementation2SpecMap getImplementation2SpecMap(){
	return specMap;
    }

    private void setImplementation2SpecMap(Implementation2SpecMap map){
	specMap = map;
    }

    /**
     * creates a new services object containing a copy of the java info of
     * this object and a new TypeConverter (shallow copy)
     * @return the copy
     */
    public Services copy() {
	Debug.assertTrue
	    (!(getJavaInfo().getKeYProgModelInfo().getServConf() 
	       instanceof SchemaCrossReferenceServiceConfiguration),
	     "services: tried to copy schema cross reference service config.");
	Services s = new Services
	    (getJavaInfo().getKeYProgModelInfo().getServConf(),
	     getJavaInfo().getKeYProgModelInfo().rec2key().copy());
	s.setTypeConverter(getTypeConverter().copy(s));
	s.setImplementation2SpecMap(specMap.copy(s));
	s.setExceptionHandler(getExceptionHandler());
	s.setNamespaces(namespaces.copy());
	return s;
    }

    /**
     * creates a new service object with the same ldt information 
     * as the actual one
     */
    public Services copyPreservesLDTInformation() {
	Debug.assertTrue
	    (!(javainfo.getKeYProgModelInfo().getServConf() 
	       instanceof SchemaCrossReferenceServiceConfiguration),
	     "services: tried to copy schema cross reference service config.");
	Services s = new Services(getExceptionHandler());
	s.setTypeConverter(getTypeConverter().copy(s));
	s.setImplementation2SpecMap(specMap.copy(s));
	s.setNamespaces(namespaces.copy());
	return s;
    }
    
    public Services copyProofSpecific() {
        Services s = new Services(getJavaInfo().getKeYProgModelInfo().getServConf(),
                getJavaInfo().getKeYProgModelInfo().rec2key());
        s.setTypeConverter(getTypeConverter().copy(s));
        s.setImplementation2SpecMap(specMap.copy(s));
        s.setExceptionHandler(getExceptionHandler());
        s.setNamespaces(namespaces.copy());
        return s;
    }

    /*
     * returns an existing named counter, creates a new one otherwise
     */
    public Counter getCounter(String name) {
        Counter c = (Counter) counters.get(name);
        if (c!=null) return c;
        c = new Counter(name);
        counters.put(name, c);
        return c;
    }
    
    public void setBackCounters(Node n) {
        Iterator it = counters.values().iterator();
        while (it.hasNext()) ((Counter)it.next()).undo(n);
    }
    
    /**
     * returns the namespaces for functions, predicates etc.
     * @return the proof specific namespaces
     */
    public NamespaceSet getNamespaces() {
        return namespaces;
    }
    
    /**
     * sets the namespaces of known predicates, functions, variables
     * @param namespaces the NamespaceSet with the proof specific namespaces
     */
    public void setNamespaces(NamespaceSet namespaces) {
        this.namespaces = namespaces;
    }
}
