// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.logic.ldt;

import de.uka.ilkd.key.java.abstraction.ListOfType;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.proof.RuleSource;

/** This class inherits from LDT and implements all method that are
 * necessary to handle integers, shorts and bytes.
 * It loads the needed rule file and offers method to convert java
 * number types to their logic counterpart
 */
public class IntegerLDT extends NumberLDT {
    
    public IntegerLDT() {
	super(new Name("integer"));
        ruleSource = RuleSource.initRuleFile("intRules.key");
    }


    /** returns the name of the target sort of the LDT 
     * is needed to find the right Sort when constructing the LDT
     * @return the Name of the target sort
     */
    public Name getSortName() {
	return targetSort() == null ? new Name("int") : targetSort().name();
    }

    public void init(Namespace func_ns, Sort targetSort, ListOfType types) {
	super.init(func_ns, targetSort, types);
	// the operators

	plus            = addFunction((Function)func_ns.lookup(new Name("add")));
	minus           = addFunction((Function)func_ns.lookup(new Name("sub")));
	negative        = addFunction((Function)func_ns.lookup(new Name("neg")));
	negativeNumber  = addFunction((Function)func_ns.lookup(new Name(NEGATIVE_LITERAL_STRING)));
	times           = addFunction((Function)func_ns.lookup(new Name("mul")));
	divide          = addFunction((Function)func_ns.lookup(new Name("jdiv")));
	modulo          = addFunction((Function)func_ns.lookup(new Name("jmod")));	
	unaryMinusJint  = addFunction((Function)func_ns.lookup(new Name("unaryMinusJint")));	
	unaryMinusJlong = addFunction((Function)func_ns.lookup(new Name("unaryMinusJlong")));	
	addJint         = addFunction((Function)func_ns.lookup(new Name("addJint")));	
	addJlong        = addFunction((Function)func_ns.lookup(new Name("addJlong")));	
	subJint         = addFunction((Function)func_ns.lookup(new Name("subJint")));	
	subJlong        = addFunction((Function)func_ns.lookup(new Name("subJlong")));	
	mulJint         = addFunction((Function)func_ns.lookup(new Name("mulJint")));	
	mulJlong        = addFunction((Function)func_ns.lookup(new Name("mulJlong")));	
	modJint         = addFunction((Function)func_ns.lookup(new Name("modJint")));	
	modJlong        = addFunction((Function)func_ns.lookup(new Name("modJlong")));	
	divJint         = addFunction((Function)func_ns.lookup(new Name("divJint")));	
	divJlong        = addFunction((Function)func_ns.lookup(new Name("divJlong")));	
	shiftrightJint  = addFunction((Function)func_ns.lookup(new Name("shiftrightJint")));	
	shiftrightJlong = addFunction((Function)func_ns.lookup(new Name("shiftrightJlong")));	
	shiftleftJint   = addFunction((Function)func_ns.lookup(new Name("shiftleftJint")));	
	shiftleftJlong  = addFunction((Function)func_ns.lookup(new Name("shiftleftJlong")));	
	unsignedshiftrightJint  = 
	    addFunction((Function)func_ns.lookup(new Name("unsignedshiftrightJint")));	
	unsignedshiftrightJlong = 
	    addFunction((Function)func_ns.lookup(new Name("unsignedshiftrightJlong")));	
	orJint          = addFunction((Function)func_ns.lookup(new Name("orJint")));	
	orJlong         = addFunction((Function)func_ns.lookup(new Name("orJlong")));	
	andJint         = addFunction((Function)func_ns.lookup(new Name("andJint")));	
	andJlong        = addFunction((Function)func_ns.lookup(new Name("andJlong")));	
	xorJint         = addFunction((Function)func_ns.lookup(new Name("xorJint")));	
	xorJlong        = addFunction((Function)func_ns.lookup(new Name("xorJlong")));	
	negJint         = addFunction((Function)func_ns.lookup(new Name("negJint")));	
	negJlong        = addFunction((Function)func_ns.lookup(new Name("negJlong")));
		

	// predicates to test if in certain type bounds
	addFunction((Function)func_ns.lookup(new Name("inByte")));
	addFunction((Function)func_ns.lookup(new Name("inShort")));
	addFunction((Function)func_ns.lookup(new Name("inInt")));
	addFunction((Function)func_ns.lookup(new Name("inLong")));
	addFunction((Function)func_ns.lookup(new Name("inChar")));

    }
    
    /** returns true if the LDT offers an operation for the given java
     * operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param subs the logic subterms of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterms 
     */
    public boolean isResponsible
	(de.uka.ilkd.key.java.expression.Operator op, Term[] subs) {
	if (subs.length == 1) {
	    return isResponsible(op, subs[0]);
	} else if (subs.length == 2) {
	    return isResponsible(op, subs[0], subs[1]);	
	}
	return false;	
    }

    /** returns true if the LDT offers an operation for the given
     * binary java operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param left the left subterm of the java operator
     * @param right the right subterm of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterms 
     */
    public boolean isResponsible
	(de.uka.ilkd.key.java.expression.Operator op, Term left, Term right) {
	if (left!=null && left.sort() == targetSort() 
	    && right!=null && right.sort() == targetSort()) {
	    if (getFunctionFor(op) != null) {
		return true;
	    }
	}
	return false;
    }

    /** returns true if the LDT offers an operation for the given
     * unary java operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param sub the logic subterms of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterm
     */
    public boolean isResponsible
	(de.uka.ilkd.key.java.expression.Operator op, Term sub) {
	if (sub.sort() == targetSort()) {
	    if (op instanceof Negative) {
		return true;
	    }
	}
	return false;
    }

} 
