// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op;

import java.util.HashSet;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.util.Debug;

public class ModalOperatorSV extends SchemaVariableAdapter {

    private int arity;
    private HashSet modalities;
    
    /** creates a new SchemaVariable. That is used as placeholder for
     * modal operators.
     * @param name the Name of the SchemaVariable
     * @param arity the arity of the modal operators matched by this SV
     * @param modalities modal operators matched by this SV
     */    
    ModalOperatorSV(Name name, int arity, HashSet modalities) {
        super(name, Modality.class);
	this.arity = arity;
	this.modalities = modalities;
    }
    

    /**
     * checks whether the top level structure of the given @link Term
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.  
     * @return true iff the top level structure of
     * the {@link Term} is valid.
     */
    public boolean validTopLevel(Term term){
	if (term.arity() != this.arity) return false;
	boolean result = true;
	for(int i=0;i<term.arity(); i++)
	    result = result && term.sub(i).sort().equals(Sort.FORMULA);
	return result;
    }

    /**
     * determines the sort of the {@link Term} if it would be created using this
     * Operator as top level operator and the given terms as sub terms. The
     * assumption that the constructed term would be allowed is not checked.
     * @param term an array of Term containing the subterms of a (potential)
     * term with this operator as top level operator
     * @return sort of the term with this operator as top level operator of the
     * given substerms
     */
    public Sort sort(Term[] term){
	return Sort.FORMULA;
    }


   /** @return arity of the Operator as int */
    public int arity(){
	return this.arity;
    }

    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be
     * changed by modalities
     */
    public boolean isRigid (Term term) {
	return (term.javaBlock () == JavaBlock.EMPTY_JAVABLOCK);
    }

    
    
    /**
     * (non-Javadoc)
     * @see de.uka.ilkd.key.logic.op.Operator#match(Operator, de.uka.ilkd.key.rule.MatchConditions, de.uka.ilkd.key.java.Services)
     */
    public MatchConditions match(Operator op, MatchConditions mc,
            Services services) {        
        if (!(op instanceof Modality)) {
            Debug.out("FAILED. ModalitySV matches only modalities (template, orig)",
                    this, op);
            return null;
        }  
        final Modality modalOp = (Modality)op;
        if(modalities().contains(modalOp)) {
            Modality o=(Modality) mc.getInstantiations().getInstantiation(this);
            if (o == null) {
                return mc.
                  setInstantiations(mc.getInstantiations().add(this, modalOp));
            } else {
                if(o != modalOp) {
                    Debug.out("FAILED. Already instantiated with a different modality.");
                    return null;
                }
            }                           
	}
        Debug.out("FAILED. template is schema modal,"
                +" term is a modality, but not a matching one");
        return null; // template is schema modal, term is a modality, but not a matching one     
    }
    // SchemaVariable methods

    public boolean isModalOperatorSV() {
	return true;
    }
    
    public HashSet modalities() {
	return modalities;
    }
     
    /** toString */
    public String toString() {
	return toString(" (modal operator)");
    }

}
