// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator.heuristics;

/**
 * Implements a simple heuristic for the if-then-else-construct:
 * if the if- and else- part are both boolean expression, we don't
 * have to invent a new function symbol necessarily.
 * In this case we can use a formula (or boolean term) directly.
 * This heuristic specifies that all such occurrences of if-then-else
 * should be treated by a formula.
 */
import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.*;

import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;


public class IfThenElseHeuristic extends KeyTranslationHeuristic {
    
    /**
     * @return the name of this translation heuristic.
     */
    public String getHeuristicName(){
	return "If-Then-Else Optimization - Heuristic";
    }
    
    // ================ Start the computation of the translation types =====================
	    
    public void determineTranslationTypesFor(OclTree constraintTree, 
					     AExpression expression, 
					     Map currentMarking,
					     KeyOclExpressionTranslator exprTranslator){
		
	this.tree = constraintTree;
	this.expression = expression;
	this.exprTranslator = exprTranslator;

	this.varMap = new HashMap();
	nodeTranslationTypeMap = currentMarking;
	
	expression.apply(this);

	
    }
  

    // ================ The main part of the heuristic ===================================    
    // Use a formula representation for if-the-else-expresssions with 
    // boolean expressions for each branch 

     protected void translateIfThenElseExpression(Node ie, Node ifExpression, Node thenExpression,
						 Node elseExpression){
	log("Translating if-then-else-construct.");
	

	Type thenType = tree.getNodeType(thenExpression);
	Type elseType = tree.getNodeType(elseExpression);
	
	
	if(thenType.equals(Basic.BOOLEAN) && elseType.equals(Basic.BOOLEAN) ){
	    // The single cases are all boolean valued ...
	    // Therefore we can translate the if-feature mor directly without
	    // introducing a new symbol 'if-k':
	    // We can simply use the formula: 
	    //    (transIfExp -> transThenExp) & (!transIfExp -> transElseExp)
	
	    // So we mark the corresponding node to be represented by a formula
	    // (instead of a term and some axioms!)

	    TranslationType currentType =  getTranslationTypeForNode(ie);
	    currentType.addMark(TranslationMark.AS_PREDICATE);
	    setTranslationTypeForNode(ie, currentType);

	    log("Marked node as predicative :" + ie.toString());
	}

    }
    
}








