// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator.test;

import tudresden.ocl.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.pp.*;
import de.uka.ilkd.key.*;
import java.io.*;
import java.net.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;

/**
 * Performs the translation for OCL constrains as specified
 * in a tranformation test file.
 * The results of the execution wille be logged in a file.
 */


public class TranslationTester {

    TestSuiteReader testSuiteReader;
    FileWriter fileWriter;

    KeyOclConstraintTranslator constraintTranslator;
    DefaultTranslationControl tControl;

    tudresden.ocl.check.types.ModelFacade mf;
    ProverFacade proverFacade;
    boolean enableLog;
    boolean translationDetails;

    public TranslationTester(){
	testSuiteReader = new TestSuiteReader();
    }
    

    public void executeTestSuite(String testSuiteFileName, String testResultFileName){
	// First read the test suite file ...
	System.err.println("\n\n\n\n==========================================================================" );
	System.err.println("================= Executing a test suite .... ============================" );
	System.err.println("==========================================================================" );
	

	System.err.print("Reading test suite from file : " + testSuiteFileName + " ..." );
	testSuiteReader.readTestSuite(testSuiteFileName);
	System.err.println("... finished !" );

	System.err.print("Getting test suite ..." );
	TransformationTestSuite testSuite = testSuiteReader.getTestSuite();
	System.err.println("... finished !" );	

	
	System.err.println("\nExecuting test suite, logging results in file : " + testResultFileName );
	String suiteName = testSuite.getName();
	if (suiteName.length() == 0){ suiteName = "UNNAMED-TEST-SUITE";}
	System.err.println("\n\nTestSuite name : " + suiteName);

	String description = testSuite.getDescription();
	System.err.println("TestSuite description : " + description);
	// Now we can execute the test cases ...
	try {
	    fileWriter = new FileWriter(testResultFileName);
	    String s;
	    
	    String modelFile = testSuite.getXMIModelFile();
	    
	    s = "---------------------------------------------------------------------------------------\n";
	    s += "------------------------    Executing a test suite ....     ---------------------------\n";
	    s += "---------------------------------------------------------------------------------------\n\n\n";
	    s += "Executing test suite in file : " + testSuiteFileName + "\n";
	    s += "Testsuite name : " + suiteName + "\n";
	    s += "\nDescription :\n";
	    s += description;
	    s += "\n\n";
	    s += "XMI model file : " + modelFile + "\n";
	    s += "\n";

	    System.err.println("\n\nReading XMI model file ...");
	    
	    File f = new File(modelFile);
	    
	    URL xmiFileURL = f.toURL();
	    

	    try {
		mf = tudresden.ocl.check.types.xmifacade.XmiParser.createModel(xmiFileURL, xmiFileURL.toString());
	    } catch (Exception e) {
		System.err.print(" ... an ERROR occured while reading the XMI model file !");
		e.printStackTrace();
		return;
	    }
	    System.err.println(" ... finished !\n\n");

	   
	   
	    s += "---------------------------------------------------------------------------------------\n";
	    s += "\nExecuting the single test cases in the test suite ...\n\n";
	    
	    fileWriter.write(s);
	    fileWriter.flush();

	    
	    TransformationTestCase[] testCases = testSuite.getTestCases();
	    TransformationTestCase currentTestCase;
	    boolean ok;
	    String testName;
	   
	    for (int i=0; i<testCases.length;i++){
		
		
		// Create a new Casetool-Facade object ...
		
		// System.err.println("\nCreating new facade object ........................................");
		try {
		    proverFacade = new TestFacade
			(testSuite.getImplementationSourcesDirectory()); // new TestFacade();
		    
		} catch (IOException e){
		    // this is not of interest here, since the Taclets are not used !
		    e.printStackTrace();
		}
		// System.err.println(".......................................................... finished !\n");


		currentTestCase = testCases[i];
		testName = currentTestCase.getName();
		if (testName.length() == 0){ testName = "UNNAMED-TEST-CASE";}
		System.err.print("[ " + (i+1) + " ]  Execute new testcase : " + testName + " -------- ");
		ok = executeTestCase(currentTestCase);
		if (ok){
		    System.err.println("  ... ok!   :-)");
		} else {
		    System.err.println("  ... an error occured !   :-( \n"); 
		}
		
	    }

	   
	    fileWriter.flush();

	    fileWriter.close();

	     
	} catch (IOException ioe){
	    ioe.printStackTrace();
	}

	System.err.println("\n\n... finished !"); 
	System.err.println("\nPlease look at '" + testResultFileName + "' for a textual representation of the results of the translation." );
    }


    public boolean executeTestCase(TransformationTestCase testCase) throws IOException {
	boolean result = true;
	
	String name = testCase.getName();
	if (name.length() == 0){ name = "UNNAMED-TEST-CASE";}
	String description = testCase.getDescription();
	if (description.length() == 0){ description = "NO-DESCRIPTION";}
	String constraintText = testCase.getConstraint();
	// Handle the specified options for the test case ...

	// Currently we only handle the diamond form option !
	boolean useDiamond = false;
	String diamondOptionValue = testCase.getValueForOption("diamond-form");
	if (diamondOptionValue != null && diamondOptionValue.equals("true")){
	    useDiamond = true;
	}
	
	boolean translationControllClassSpecified = false;
	String translationControlClassValue = testCase.getValueForOption("translation-control-class");
	DefaultTranslationControl transController = null;
	String className = "";
	if (translationControlClassValue != null) {
	    translationControllClassSpecified = true;
	    className = translationControlClassValue;
	    try {
		Class translationControllerClass = Class.forName(className);
		Object translationControllerObject =  translationControllerClass.newInstance();
		transController  = (DefaultTranslationControl) translationControllerObject;
	    } catch (Exception e){
		e.printStackTrace();
		return false;
	    }
	}
	

	boolean isExpression = testCase.isOCLExpression();
	
	// The options of a testcase  are currently not used here !!!
	
	String s = "";
	
	s += "---------------------------------------------------------------------------------------\n";
	s += "Executing next test case ...\n";
	s += "---------------------------------------------------------------------------------------\n";
	s += "Testcase name : " + name +"\n";
	s += "\nTestcase description :\n";
	s += description;
	s += "\n\n";
	if (useDiamond){
	    s += "- Translation option: Using diamond form for result formula !"  +"\n";
	}
	if (isExpression){
	    s += "- Current test case tests an OCL expression (instead of an OCL constraint) !"  +"\n";
	}
	if (translationControllClassSpecified){
	    s += "- Using a special translation control object of class : "+ className  +" !\n";
	}
	
	
	
	s += "\nConstraint to be translated : \n";
	s += constraintText;
	
	s += "\n\nTranslation result : \n\n";
	
	try {
	    KeyTranslationResult translationResult =  
		performTranslationFor(constraintText, isExpression, useDiamond, transController);
	    if (isExpression == false){
		
		// NEW: Use a pretty printed version of the translation result ...
		// Since there is currently no explicit pretty printer for formulas, 
		// we use the pretty printer for sequents ...
		Term resultTerm = translationResult.getTerm();
		Semisequent semiSeq =  Semisequent.
		    EMPTY_SEMISEQUENT.insertFirst(new ConstrainedFormula(resultTerm)).semisequent();

		Sequent seq = Sequent.createSuccSequent(semiSeq);
		
		
		ProgramPrinter prgPrinter = new ProgramPrinter(new StringWriter());
		LogicPrinter seqPrinter = new LogicPrinter(prgPrinter,
							   NotationInfo.createInstance(),
							   true);
		seqPrinter.printSequent(seq);
		String resultString = seqPrinter.toString();
		
		s += resultString + "\n\n\n\n";


		
	    } else {
		s += translationResult.toString() + "\n\n\n\n";
	
	    }
	} catch (Exception e){
	    s += "#######################################################################################\n"; 
	    s += "An ERROR occured during translation !!!!!!!!!!!!!\n\n";
	    s += e.toString() + "\n"; 
	     s += "#######################################################################################\n\n\n\n"; 
	    e.printStackTrace();
	    result = false;
	}
	// s += "---------------------------------------------------------------------------------------\n";
	fileWriter.write(s);
	fileWriter.flush();

	return result;
	
    }

    public KeyTranslationResult performTranslationFor(String constraintText,boolean isExpression, 
						      boolean useDiamond,
						      DefaultTranslationControl tControl) throws java.io.IOException {

	// First create an OclTree object for the constraint ...
	OclTree constraintTree;
	if (isExpression == false) {
	    constraintTree = OclTree.createTree(constraintText, mf, new KeyTypeCheckerFactory());
	} else {
	    // We only want to translate an arbitrary OCL expression instead of a constraint
	    // Therefore we have to modify the type checking for top level expressions in constraints ...
	    constraintTree = OclTree.createTree(constraintText, mf, new KeyTypeCheckerFactory(true));
	}
	
	KeyTranslationResult aRes;
	
	boolean diamondForm = useDiamond; 

	// This should be handled with the OPTION part of the testcase ...
	boolean ignorePost = false;
	
	constraintTranslator = new KeyOclConstraintTranslator(proverFacade);
	if (tControl != null){
	    constraintTranslator.setTranslationControl(tControl);
	}
	
	if (this.translationDetails){
	    constraintTranslator.setLoggingEnabled(true);
	}
        aRes = 
	    (KeyTranslationResult) constraintTranslator.getTranslationFor(constraintTree, 
									  diamondForm, ignorePost,
									  isExpression, Op.DIA);
	return aRes;

    } 

    public void enableLogging(boolean logOn){
	this.enableLog = logOn;
	if (logOn) {
	    testSuiteReader.enableOutputLogging(System.err);
	} else {
	    testSuiteReader.disableOutputLogging();
	}
    }
    
    
    public void showTranslationDetails(boolean transDetails){
	this.translationDetails = transDetails;
    } 

    public static void main(String[] args){
	if (args.length == 2){
	    String inputFileName = args[0];
	    String outputFileName = args[1];
	    TranslationTester aTransTester = new TranslationTester();
	    aTransTester.executeTestSuite(inputFileName, outputFileName);
	} else if (args.length == 3){
	    String options =  args[0];
	    String inputFileName = args[1];
	    String outputFileName = args[2];
	    
	    boolean logParsing = false;
	    boolean localTranslationDetails = false;
	    if (options.indexOf("v") != -1){
		logParsing = true;
	    }

	    if (options.indexOf("d") != -1){
		localTranslationDetails = true;
	    }
	    
	    TranslationTester aTransTester = new TranslationTester();
	    aTransTester.enableLogging(logParsing);
	    aTransTester.showTranslationDetails(localTranslationDetails);
	    aTransTester.executeTestSuite(inputFileName, outputFileName);
	    
	} else {
	    System.err.println("Wrong syntax !\nUse 'java de.uka.ilkd.key.ocl.ocldltranslation.TranslationTester [-v] testSuiteFileName  resultFileName'");
	    System.err.println("Arguments :");
	    System.err.println("-----------");
	    System.err.println("testSuiteFileName : The file that contains all informations for the translation.");
	    System.err.println("resultFileName : The file in which the results of the translation will be witten.\n");
	    System.err.println("Options :");
	    System.err.println("---------");
	    System.err.println("v : Show details of the XML parsing process for the test suite file.");
	    // System.err.println("d : Show details of the translation process for the test cases.");
	}

    }


}


