// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.collection.ListOfString;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.SortedSchemaVariable;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.rule.TacletApp;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * Proposes names for variables (except program variables).
 */
public class VariableNameProposer implements InstantiationProposer {

    /**
     * An instance of VariableNameProposer.
     */
    public static final VariableNameProposer DEFAULT 
    				= new VariableNameProposer();

    private static final String DEPENDING_VARIABLE_NAME_POSTFIX = "_";
    private static final String VARIABLE_NAME_PREFIX            = "_var";
    private static final String LABEL_NAME_PREFIX               = "_label";

    private static final String DEPENDINGVARCOUNTER_PREFIX = "DepVarCnt";
    private static final String VARCOUNTER_NAME 	   = "VarCnt";
    private static final String LABELCOUNTER_NAME 	   = "LabelCnt";

    /**
     * HACK.
     */
    private static HashMap globalDependingVariableNameCnt = new HashMap();


    /**
     * Returns an instantiation proposal for the schema variable var.
     * Currently supports names for depending SVs, variable SVs, and labels.
     */
    public String getProposal(TacletApp app,
    			      SchemaVariable var,
			      Services services,
			      Node undoAnchor,
			      ListOfString previousProposals) {
	if(var.isDependingSV()) {
	    return getNameProposalForDependingVariable(app,
	    					       var,
						       services,
						       undoAnchor);
	} else if(var.isVariableSV()) {
	    return getNameProposalForVariableSV(app,
	    					var,
						services,
						undoAnchor);
	} else if(((SortedSchemaVariable)var).sort() == ProgramSVSort.LABEL) {
	    return getNameProposalForLabel(app,
	    				   var,
					   services,
					   undoAnchor);
	} else {
	    return null;
	}
    }


    /**
     * Generates a proposal for the instantiation of the given term
     * schema variable, which is declared as depending (i.e. skolem
     * function).
     */
    private String getNameProposalForDependingVariable(TacletApp p_app,
    						       SchemaVariable p_var,
						       Services services,
						       Node undoAnchor) {
	return getNameProposalForDependingVariable
	    ( createBaseNameProposalBasedOnCorrespondence ( p_app, p_var ),
	      services,
	      undoAnchor );
    }


    /**
     * Find a name for the variable <code>p_var</code>, based on the result
     * of <code>Taclet.getNameCorrespondent</code>
     */
    protected static String createBaseNameProposalBasedOnCorrespondence (TacletApp p_app,
                                                                         SchemaVariable p_var) {
        final String result;
        final SchemaVariable v = p_app.taclet ().getNameCorrespondent ( p_var );
        if ( v != null && p_app.instantiations ().isInstantiated ( v ) ) {
            result = "" + p_app.instantiations ().getInstantiation ( v );
        } else {
            // ... otherwise use the name of the dependingSV
            result = "" + p_var.name ();
        }

        // remove characters that should better not turn up in identifiers
        // more or less a HACK
        final Pattern pattern = Pattern.compile ( "[^_#a-zA-Z0-9]" );
        final Matcher matcher = pattern.matcher ( result );

        return matcher.replaceAll ( "_" );
    }


    private String getNameProposalForDependingVariable(String name,
    						       Services services,
						       Node undoAnchor) {
	name += DEPENDING_VARIABLE_NAME_POSTFIX;
    name += services.getCounter(DEPENDINGVARCOUNTER_PREFIX + name)
			.getCountPlusPlusWithParent(undoAnchor);
	return name;
    }


    /**
     * HACK - DO NOT USE.
     * Called solely from class rule/inst/IntroduceNewAnonymousUpdate.
     * Use getProposal() instead.
     */
    public String getNameProposalForDependingVariable(String name) {
	name += DEPENDING_VARIABLE_NAME_POSTFIX;

	synchronized ( globalDependingVariableNameCnt ) {
	    Integer cnt = (Integer)globalDependingVariableNameCnt.get (name);

	    if ( cnt == null )
		cnt = new Integer ( 0 );
	    else
		cnt = new Integer ( cnt.intValue () + 1 );

	    globalDependingVariableNameCnt.put(name, cnt);

	    name += cnt;
	}
	return name;
    }


    /**
     * Generates a proposal for the instantiation of the given
     * schema variable, which is a variable SV.
     */
    private String getNameProposalForVariableSV(TacletApp app,
						SchemaVariable var,
						Services services,
						Node undoAnchor) {
	return VARIABLE_NAME_PREFIX + services.getCounter(VARCOUNTER_NAME)
	  				      .getCountPlusPlusWithParent(undoAnchor);
    }


    /**
     * Generates a proposal for the instantiation of the given
     * schema variable, which is of sort label.
     */
    private String getNameProposalForLabel(TacletApp app,
					   SchemaVariable var,
					   Services services,
					   Node undoAnchor) {
	return LABEL_NAME_PREFIX + services.getCounter(LABELCOUNTER_NAME)
	    				   .getCountPlusPlusWithParent(undoAnchor);

    }
}
