// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;

import de.uka.ilkd.key.gui.LibrariesSettings;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.Choice;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.NamespaceSet;
import de.uka.ilkd.key.logic.SetOfChoice;
import de.uka.ilkd.key.parser.KeYParser;
import de.uka.ilkd.key.parser.KeYLexer;
import de.uka.ilkd.key.parser.ParserConfig;
import de.uka.ilkd.key.parser.ParserMode;
import de.uka.ilkd.key.proof.CountingBufferedInputStream;
import de.uka.ilkd.key.proof.ProblemLoader;
import de.uka.ilkd.key.proof.RuleSource;
import de.uka.ilkd.key.proof.mgt.Contract;
import de.uka.ilkd.key.proof.mgt.Contractable;
import de.uka.ilkd.key.rule.SetOfTaclet;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ProgressMonitor;

/** represents an input from a .key problem file producing environment as
 * well as a proof obligation.
 */
public class KeYFile implements EnvInput {

    /** the RuleSource delivering the input stream for the file.
     */
    protected RuleSource file = null;

    /** maps methods to their jml specifications. */ 
    protected HashMap method2jmlspecs=null;

    
    protected InputStream input = null;

    /** the initial configuration to be used (and modified) while reading.
     */
    protected InitConfig initConfig = null;

    private String name;

    protected KeYParser problemParser=null;
    
    protected ProofSettings settings;

    /** the graphical entity to notify on the state of reading.
     */
    protected ProgressMonitor monitor;

    /** creates a new representation for a given file by indicating a name
     * and a file representing the physical source of the .key file.
     */
    public KeYFile(String name, File file) {
	this(name, RuleSource.initRuleFile(file), null);
    }

    /** creates a new representation for a given file by indicating a name
     * and a file representing the physical source of the .key file.
     * @param monitor the progress monitor to notify on the state of reading
     */
    public KeYFile(String name, File file, ProgressMonitor monitor) {
	this(name, RuleSource.initRuleFile(file), monitor);
    }

    /** creates a new representation for a given file by indicating a name
     * and a RuleSource representing the physical source of the .key file.
     * @param monitor the progress monitor to notify on the state of reading
     */
    public KeYFile(String name, RuleSource file, ProgressMonitor monitor) {
	this.file=file;
	this.name=name;
	this.monitor=monitor;
    }

    /** returns the initial configuration that is used to read and
     * that may be modified by reading.
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** returns the number of chars in the .key file.
     */
    public int getNumberOfChars() {
	return file.getNumberOfChars();
    }
    

    protected InputStream getNewStream() 
    throws FileNotFoundException {
        try {
            if (input != null) input.close();
        } catch(IOException ioe) {
            System.err.println("WARNING: Cannot close stream "+file+"\n"+ioe);
        }
        if (!file.isAvailable()) {
          throw new FileNotFoundException("File/Resource " + file + " not found.");  
        } 
        input = file.getNewStream();
        return input;
    }
    

    /** reads the whole .key file and modifies the initial configuration
     * assigned to this object according to the given modification strategy. 
     * Throws an exception if no initial configuration has been set yet.
     */
    public void read(ModStrategy mod) throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	try {
            Debug.out("Reading KeY file", file);
	    CountingBufferedInputStream cinp = 
		new CountingBufferedInputStream
		    (getNewStream(),monitor,getNumberOfChars()/100);
	    ParserConfig pc = new ParserConfig
		(initConfig.getServices(), 
		 initConfig.namespaces().copy(), 
		 initConfig.getDefaultChoice().copy());
	    problemParser = new KeYParser
		(ParserMode.PROBLEM, new KeYLexer(cinp,initConfig.getServices().getExceptionHandler()), file.toString(), pc, pc, initConfig.
		 getTaclet2Builder(), initConfig.getTaclets(), initConfig.getActivatedChoices()); 
            problemParser.problem(); 
	    initConfig.setDefaultChoice(problemParser.getDefaultChoice());
	    initConfig.addCategory2DefaultChoices(problemParser.
						  getCategory2Default());
	    SetOfTaclet st = problemParser.getTaclets();
	    initConfig.setTaclets(st);
	    initConfig.add(problemParser.namespaces(), mod);
	    if (initConfig.getProofEnv()!=null) {
		initConfig.getProofEnv().addContracts
		    (problemParser.getContracts(), null);
	    }
            Debug.out("Read KeY file   ", file);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	} catch (FileNotFoundException fnfe) {
	    throw new ProofInputException(fnfe);
        }
    }

    public Includes readIncludes() throws ProofInputException{
	if(file == null) return new Includes();
	try {
	    ParserConfig pc = new ParserConfig
		(new Services(), 
		 new NamespaceSet(), 
		 new Choice(new Name("Default")));
            // FIXME: there is no exception handler here, thus, when parsing errors are ecountered
	    // during collection of includes (it is enough to mispell \include) the error
	    // message is very uninformative - ProofInputException without filename, line and column
	    // numbers. Somebody please fix that. /Woj
	    problemParser = new KeYParser
		(ParserMode.PROBLEM, new KeYLexer(getNewStream(),null), file.toString(), 
		 pc, pc, null, null, null); 
            problemParser.parseIncludes(); 
	    return problemParser.getIncludes();
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
        } catch (FileNotFoundException fnfe) {
            throw new ProofInputException(fnfe);
        } catch(de.uka.ilkd.key.util.ExceptionHandlerException ehe){
	    throw new ProofInputException(ehe.getCause().getMessage());
	}
    }


    public LibrariesSettings readLibrariesSettings() throws ProofInputException{
        if (settings == null) {
            getPreferences();
        }
        if (settings == null || settings.getLibrariesSettings().emptyProperties()) {
            return ProofSettings.DEFAULT_SETTINGS.getLibrariesSettings();
        }
        return settings.getLibrariesSettings();
    }
    
    
    public SetOfChoice readActivatedChoices() throws ProofInputException{
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	try {
	    ParserConfig pc = new ParserConfig
		(initConfig.getServices(), 
		 initConfig.namespaces(), 
		 initConfig.getDefaultChoice());
	    problemParser = new KeYParser
		(ParserMode.PROBLEM, new KeYLexer(getNewStream(),
		        initConfig.getServices().getExceptionHandler()), 
		        file.toString(), pc, pc, null, null, null); 
	    problemParser.parseWith();
	    return	(problemParser.getActivatedChoices());
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);	   
	} catch (FileNotFoundException fnfe) {
            throw new ProofInputException(fnfe);
        }
    }

    private KeYParser createDeclParser() throws FileNotFoundException {
	return new KeYParser(ParserMode.DECLARATION,
	                     new KeYLexer(getNewStream(),
                                          initConfig.getServices().getExceptionHandler()),
			     file.toString(), 
			     initConfig.getServices().copy(),
			     initConfig.namespaces().copy(),
			     initConfig.getDefaultChoice());
    }

    /** reads the sorts declaration of the .key file only, 
     * modifying the sort namespace
     * of the initial configuration if allowed in the given 
     * modification strategy.
     */
    public void readSorts(ModStrategy mod) throws ProofInputException {
	try {
	    KeYParser p=createDeclParser();
	    p.parseSorts();
	    initConfig.addCategory2DefaultChoices(p.getCategory2Default());
	    initConfig.add(p.namespaces(), mod);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	} catch (FileNotFoundException fnfe) {
            throw new ProofInputException(fnfe);
        }
    }
    
    /** reads the functions and predicates declared in the .key file only, 
     * modifying the function namespaces of the respective taclet options. 
     */
    public void readFuncAndPred(ModStrategy mod) throws ProofInputException {
	if(file == null) return;
	try {
	    KeYParser p=createDeclParser();
	    p.parseFuncAndPred();
	    initConfig.add(p.namespaces(), mod);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	} catch (FileNotFoundException fnfe) {
            throw new ProofInputException(fnfe);
        }
    }

   /** reads the rules and problems declared in the .key file only, 
     * modifying the set of rules 
     * of the initial configuration if allowed in the given 
     * modification strategy.
     */
    public void readRulesAndProblem(ModStrategy mod)
    throws ProofInputException {
        
	/*
	  two namespace sets which share all namespace except the
	  variable namespace 	     	      
	*/
	NamespaceSet schema = initConfig.namespaces().copy();
	// some namespaces have to be shared
	NamespaceSet normal = new NamespaceSet();
	normal.setRuleSets(schema.ruleSets());
	normal.setFunctions(schema.functions());
	normal.setChoices(schema.choices());
	ParserConfig schemaConfig = 
	    new ParserConfig(initConfig.getServices(), schema, 
			     initConfig.getDefaultChoice());
	ParserConfig normalConfig = 
	    new ParserConfig(initConfig.getServices(), normal, 
			     initConfig.getDefaultChoice());
	try {
	    CountingBufferedInputStream cinp = new CountingBufferedInputStream
            (getNewStream(),monitor,getNumberOfChars()/100);
            problemParser = new KeYParser(ParserMode.PROBLEM,
                                              new KeYLexer(cinp,
                                                           initConfig.getServices().
                                                           getExceptionHandler()), 
                                              file.toString(),
                                              schemaConfig,
                                              normalConfig,
                                              initConfig.getTaclet2Builder(),
                                              initConfig.getTaclets(), initConfig.getActivatedChoices()); 
                       
            problemParser.parseTacletsAndProblem();
	    initConfig.add(schemaConfig.namespaces(), mod);
	    initConfig.add(normalConfig.namespaces(), mod);
	    SetOfTaclet st = problemParser.getTaclets();
	    initConfig.setTaclets(st);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	} catch (FileNotFoundException fnfe) {
            throw new ProofInputException(fnfe);
        }
    }


    public void readProblem(ModStrategy mod) throws ProofInputException {
	readRulesAndProblem(mod);
    }


    /** defines the initial configuration that is to be used to read the .key
     * file and that the read declarations are added to. The set initial 
     * configuration is overwritten by a call to setConfig().
     */
    public void addTo(InitConfig conf) {
	this.initConfig=conf;
    }

    /** defines the initial configuration that is to be used to read the .key
     * file and that the read declarations are added to. The set initial 
     * configuration is overwritten by a call to addTo().
     */
    public void setConfig(InitConfig conf) {
	this.initConfig=conf;
    }

    /** reads a saved proof of a .key file
     */
    public void readProof(ProblemLoader prl) throws ProofInputException {
        try {
	    problemParser.proof(prl);
        } catch(antlr.ANTLRException e) {
            throw new ProofInputException(e);
        }
    }
    
    public ProofSettings getPreferences() throws ProofInputException {
        if (settings == null) {
            if (file.isDirectory()) return null;
            try {
                problemParser = new KeYParser(ParserMode.PROBLEM,
	           new KeYLexer(getNewStream(),null), file.toString());
                settings = new ProofSettings();            
                settings.loadSettingsFromString(problemParser.preferences());
            } catch (antlr.ANTLRException e) {
                throw new ProofInputException(e);
            } catch (FileNotFoundException fnfe) {
                throw new ProofInputException(fnfe);
            } catch (de.uka.ilkd.key.util.ExceptionHandlerException ehe) {
                throw new ProofInputException(ehe.getCause().getMessage());
            }
        }
        return settings;
     }

    /** returns the name of the .key file
     */
    public String name() {
	return name;
    }

    public void close() {
        try {
            if (input != null) { 
		input.close();
	    }
        } catch(IOException ioe) {
            System.err.println("WARNING: Cannot close stream "+file+"\n"+ioe);
        }
    }

    public String toString() {
	return name()+" "+file.toString();
    }
    
    public boolean equals(Object o){
	if(!(o instanceof KeYFile)) return false;
	KeYFile kf = (KeYFile) o;
	if(file != null && kf.file != null){
	    return kf.file.getExternalForm().
		equals(file.getExternalForm());
	}
	return false;
    }

    public int hashCode(){
        final String externalForm = file.getExternalForm();
        if (externalForm == null)
            return -1;
	return externalForm.hashCode();
    }

    public Contractable[] getObjectOfContract() {
        return new Contractable[0];
    }

    public boolean initContract(Contract ct) {
        return false;
    }

    public void startProtocol() {
	// do nothing
    }

}
