// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import tudresden.ocl.OclTree;
import de.uka.ilkd.key.ProverFacade;
import de.uka.ilkd.key.casetool.ModelManager;
import de.uka.ilkd.key.casetool.ReprModelClass;
import de.uka.ilkd.key.casetool.ReprModelMethod;
import de.uka.ilkd.key.casetool.StandaloneJavaModelManager;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.logic.IteratorOfNamed;
import de.uka.ilkd.key.logic.SetAsListOfChoice;
import de.uka.ilkd.key.logic.SetOfChoice;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.ocl.ocldltranslator.OCLTranslationException;
import de.uka.ilkd.key.ocl.ocldltranslator.OCLTreeToDLAggregator;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.DetailedAggregator;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.OCLMod;
import de.uka.ilkd.key.ocl.tudresden.TUDresdenParserAdapter;
import de.uka.ilkd.key.proof.ProofAggregate;
import de.uka.ilkd.key.proof.RuleSource;
import de.uka.ilkd.key.proof.mgt.*;
import de.uka.ilkd.key.util.Debug;

/** Represents the proof obligation input for the prover as an OCL 
 * constraint in textual format.
 */
public abstract class OCLProofOblInput implements ProofOblInput {

    /** The initial configuration for a prover to use for reading and modify*/
    protected InitConfig initConfig = null;
    private String name;
    private String proofHeader;

    private static ModelManager modelManager;

    public static void setModelManager(ModelManager modMan) {
	if (modelManager==null) {
	    modelManager = modMan;
	}
    }

    public static ModelManager getModelManager() {
	if (modelManager==null) {
	    setModelManager(StandaloneJavaModelManager.INSTANCE);
	}
	return modelManager;
    }

    /** Creates a new representation of an OCL input with a name describing
     * the kind of input.
     */
    public OCLProofOblInput(String name) {
	this.name=name;
    }

    /** returns an OCLTreeToDLAggregator for use in the OCL-DL-Translation.
     * @param mods Modifications to OCL constraints and OCL model used
     * to parse model extensions such as excThrown or null. If this parameter
     * has length 0 a usual OCLTODLAggregator is returned, otherwise a 
     * DetailedAggregator considering the modifications is returned.
     */
    protected OCLTreeToDLAggregator getAggregator(ProverFacade facade, 
						  OCLMod[] mods,
						  ReprModelMethod aMethRepr) {
	if (mods.length==0) {
	    return new OCLTreeToDLAggregator(facade, aMethRepr);
	}
	return new DetailedAggregator(facade, mods, aMethRepr);
    }

    protected Term generateMethodPO(ProverFacade po2prover,
				    ReprModelMethod aMethRepr,
				    String aTag,
				    String preText, String postText, 
				    Modality modality,
				    String location) {
	Term t = generateMethodPO(po2prover,
				  ProofSettings.DEFAULT_SETTINGS
				  .getOCLRefinementSettings().getOCLMods(),
				  aMethRepr, aTag, preText, postText, modality,
				  location);
	return t;
    }
    /**
     * generate proof obligation as DL text from OCL formulas.
     * @param mods Modifications to OCL constraints and OCL model used
     * to parse model extensions, such as excThrown or null.
     * @param aTag controls the form of the generated text
     *           values for aTag: purepre, prepost, prepostaspl
     * purepre: only preconditions are considered.
     * prepost: is transformed into a DL-form.
     * prepostaspl transformation into PL form where diamond is omitted
     * and @pre constructs in the postconditions are eliminated.
     * @return a proof obligation term
     *  <code>preText -&gt; &lt;m&gt; postText</code>
     * @todo document parameters and check documentation of the
     * effects of different values for aTag.
     */
    protected Term generateMethodPO(ProverFacade po2prover,
				    OCLMod[] mods,
				    ReprModelMethod aMethRepr,
				    String aTag,
				    String preText, String postText,
				    Modality modality,
				    String location) {
	Debug.out("generate Method PO");	
	String contextString = aMethRepr.getContainingClassName() + "::" + 
           aMethRepr.getCallSignature();

	TUDresdenParserAdapter aAdapter = new TUDresdenParserAdapter();
	OclTree methPOAsTree;

 	// generating proofobligation: preText -> <m> postText
	methPOAsTree= aAdapter.doParse1
	    (contextString + aTag, getModelManager(), location,
	     initConfig.getProofEnv().getJavaModel().getModelTag(),
	     "context " + contextString + 
	     " pre: " + normalizeConstraint(preText) +
	     " post: " + normalizeConstraint(postText), 
	     aMethRepr.getContainingPackage());

        OCLTreeToDLAggregator aOCLTreeToDLAggregator;
	try {
	    aOCLTreeToDLAggregator =  getAggregator(po2prover, mods, aMethRepr);
	    if (methPOAsTree!=null) {
		Term methProofObl;
                // only preconditions are considered
		if (aTag.equals("purepre")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    false, true, modality);
		    // is transformed into a DL-form
		} else if ( aTag.equals("prepost")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    true, false, modality);
		    // transformation into PL form where diamond is omitted
		    // and @pre constructs in the postconditions are eliminated
		} else if ( aTag.equals("prepostaspl")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    false, false, modality);
		} else { 
		    throw new OCLTranslationException("unexpected tag in "
						      +"generatedMethodPO: " 
						      + aTag);
		}
            	return methProofObl;
	    } else {
		throw new OCLTranslationException("generateMethodPO() failed");
	    }
	} catch(Exception e){
	    System.out.println("KeY-Error: "+ e);
	    e.printStackTrace(System.out);
            throw new OCLTranslationException("Generation of method proof "
					      +"obligation failed: "
					      +e.getMessage());
	}
    }

    /** returns the problems as generated by reading the OCL constraint
     * and translating it to DL formulas.
     */
    public abstract ProofAggregate getPO();
    
    public SetOfChoice readActivatedChoices() throws ProofInputException{
	return SetAsListOfChoice.EMPTY_SET;
    }

    public Includes readIncludes() throws ProofInputException{
	RuleSource standard = RuleSource.initRuleFile("standardRules.key");
	Includes includes = new Includes();
	includes.put("standardRules", standard);
	return includes;
    }
    

    String getProofHeader(String javaPath) {
        if (proofHeader==null) {
            proofHeader=createProofHeader(initConfig, javaPath);
        }
        return proofHeader;
    }
    
    /** creates declarations necessary to save/load proof in textual form */
    static String createProofHeader(InitConfig initConfig, String javaPath) {
        String s;
        s = "\\javaSource \""+javaPath+"\";\n\n";

        IteratorOfNamed it;

/* program sorts need not be declared and 
 * there are no user-defined sorts with this kind of PO (yes?)
        s += "sorts {\n"; // create declaration header for the proof
        it = initConfig.sortNS().getProtocolled();
        while (it.hasNext()) {
	    String n=it.next().toString();
	    int i;
	    if ((i=n.indexOf("."))<0 || 
		initConfig.sortNS().lookup(new Name(n.substring(0,i)))==null) {
		//the line above is for inner classes.
		//KeY does not want to handle them anyway...
		s = s+"object "+n+";\n";
	    }
	}
        s+="}
*/
        s+="\n\n\\programVariables {\n";
        it = initConfig.progVarNS().getProtocolled();
        while (it.hasNext()) 
           s = s+((ProgramVariable)(it.next())).proofToString();

        s+="}\n\n\\functions {\n";
        it = initConfig.funcNS().getProtocolled();
        while (it.hasNext()) {
            Function f = (Function)it.next();
            // only declare @pre-functions, others will be generated automat.
            if (f.name().toString().indexOf("@pre")!=-1) {
                s += f.proofToString();
            }
        }

        s+="}\n\n";
        return s;
    }


    /** returns the initial configuration that is used to read the OCL
     * input and that is used to be modified during reading.
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** reads the OCL constraint and modifies the initial prover configuration
     * according to the given modification strategy.
     */
    public abstract void read(ModStrategy mod) throws ProofInputException;

    public void readProblem(ModStrategy mod) throws ProofInputException {
	read(mod);
    }

    /** returns false, that is the input never asks the user which
     * environment he prefers
     */
    public boolean askUserForEnvironment() {
	return false;
    }

    /** sets the initial configuration that is used to read the OCL
     * input and that is used to be modified during reading.
     */
    public void setConfig(InitConfig conf) {
	this.initConfig=conf;
    }

    public void startProtocol() {
        initConfig.sortNS().startProtocol();
        initConfig.getDefaultChoice().funcNS().startProtocol();
        initConfig.progVarNS().startProtocol();	
    }

    /** returns the name of the OCL proof obligation input.
     */
    public String name() {
	return name;
    }

    public abstract ReprModelClass getModelClass();

    /** computes the method specifications and stores the results in the
     * SpecificationRepository which belongs to the ProofEnvironment of InitCfg
     */
    public void readSpecs(){
	// library classes specifications
	SpecExtractor spex= new OCLSpecExtractor(initConfig.getServices(), 
						 getJavaPath());
 	initConfig.getProofEnv().addContracts
	    (spex.getSpecsFromLibraries(), getProofHeader(getJavaPath()));
	
	// specifications from model
	Vector reprMethods = null;
	Iterator reprMethodsIter = null;
	ReprModelClass reprClass = null;
	Set allClasses = getModelClass().getAllClasses();		
	Iterator allClassesIter = allClasses.iterator();
	while (allClassesIter.hasNext()){
	    reprClass = (ReprModelClass) allClassesIter.next();	    
	    reprMethods = reprClass.getOps();
	    reprMethodsIter = reprMethods.iterator();
	    while(reprMethodsIter.hasNext()){
		ReprModelMethod repr 
		    = (ReprModelMethod) reprMethodsIter.next();
		Modality[] allMod = new Modality[]{Op.DIA, Op.BOX};
		for (int i=0; i<allMod.length; i++) {
		    PrePostPair prepost = ContractUtil.getPreInvPost(repr);
		    MethodContract mct = new OCLMethodContract
			(repr, prepost, allMod[i], "EnsuresPost", true);
		    initConfig.getProofEnv().addContract(mct);
		    prepost = ContractUtil.getPreInvPostInv(repr);
		    mct = new OCLMethodContract(repr, prepost, 
						allMod[i],
						"EnsuresPostAndInv",
						false);
		    initConfig.getProofEnv().addContract(mct);
		    prepost = ContractUtil.getPreInvInv(repr);
		    mct = new OCLMethodContract(repr, prepost, 
						allMod[i],
						"PreservesInv",
						false);
		    initConfig.getProofEnv().addContract(mct);
		}
		PrePostPair prepost = ContractUtil.getPreInvTout(repr);
		MethodContract mct = new OCLMethodContract
		    (repr, prepost, 
		     Op.TOUT, "PreservesThroughout", false);
		initConfig.getProofEnv().addContract(mct);
	    }	    
	}	
    }


    /** reads the underlying Java model and modifies the initial configuration
     * of this object. If the latter is not yet defined an exception is
     * thrown.
     */
    public String readJava() {
	if (initConfig==null) {
	    throw new IllegalStateException("InitConfig not set.");
	}
	Recoder2KeY r2k = new Recoder2KeY(initConfig.getServices(), 
					  initConfig.namespaces());
	String[] cus = getModelClass().getClassesInPackage();
	r2k.readCompilationUnitsAsFiles(cus);
        return getModelClass().getRootDirectory();
    }

    public String getJavaPath() {
        return getModelClass().getRootDirectory();
    }


     /**
     * Converts an empty string to "true" and keeps it unchanged otherwise.
     */
    public static String normalizeConstraint(String aText) {
        if (aText.equals("")){
            return "true";
        }else{
            return aText;
        }
    }


}
