// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.RuleSet;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.key.proof.init.ProofOblInput;
import de.uka.ilkd.key.proof.init.InitConfig;
import de.uka.ilkd.key.proof.init.ModStrategy;
import de.uka.ilkd.key.proof.init.ProofInputException;
import de.uka.ilkd.key.proof.init.Includes;
import de.uka.ilkd.key.logic.SetAsListOfChoice;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.IteratorOfNamed;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.SetOfTerm;
import de.uka.ilkd.key.logic.SetOfChoice;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.statement.Catch;
import de.uka.ilkd.key.java.statement.CatchAllStatement;
import de.uka.ilkd.key.java.statement.Desugarable;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.KeYExceptionHandler;

/**
 * represents a contract on the DynamicLogic level, that is pre-condition, program,
 * post-condition, and modifies clause. It is the starting point to create both
 * proof obligation (to ensure that the contract is valid) as well as a taclet
 * (that allows to use the contract in a proof). DLHoareTripleContracts are created from
 * UserKeYProblemFiles using the \contract section.
 *
 */
public class DLHoareTripleContract extends AbstractContract
                                   implements HoareTripleContract {

    protected Term fma;
    protected String displayName;
    private SetOfTerm modifies;
    protected Statement sta;
    protected Taclet lemma;
    private RuleSet ruleSet;
    private String name;
    protected Services services;
    protected Namespace programVars;

    public DLHoareTripleContract(Term fma, SetOfTerm modifies,
                                 Namespace heurNS, String name,
                                 String displayName,
                                 RuleSet ruleSet,
                                 Services services,
				 Namespace programVars) {
        JavaProgramElement pe = fma.sub(1).javaBlock().program();
        sta = (Statement)((NonTerminalProgramElement) pe).getChildAt(0);
        this.fma = fma;
        this.modifies = modifies;
        this.name = name;
        this.displayName = displayName;
        this.ruleSet = ruleSet==null
                        ? (RuleSet) heurNS.lookup(new Name("use_method_specs"))
                        : ruleSet;
        this.services = services;
	this.programVars = programVars;
    }


    public Term getPre() {
        return fma.sub(0);
    }

    public String getPreText(){
        return getPre().toString();
    }

    public Term getPost() {
        return fma.sub(1).sub(0);
    }

    public String getPostText(){
        return getPost().toString();
    }

    public Modality getModality() {
        return ModalityClass.getNormReprModality((Modality) fma.sub(1).op());
    }

    private Operator getSchemaModality() {
        return ModalityClass.getSchemaModality((Modality)fma.sub(1).op());
    }

    public SetOfTerm getModifies() {
        return modifies;
    }

    public String createProgramVarsSection(){
        IteratorOfNamed it = programVars.allElements().iterator();
        String result = "\\programVariables {\n";
        while(it.hasNext()){
            ProgramVariable pv = (ProgramVariable) it.next();
	    String typeName = pv.getKeYJavaType().getJavaType().getFullName();
	    // This is really annoying! /Woj
	    if("[B".equals(typeName)) typeName = "byte[]"; else
	    if("[C".equals(typeName)) typeName = "char[]"; else
	    if("[D".equals(typeName)) typeName = "double[]"; else
	    if("[F".equals(typeName)) typeName = "float[]"; else
	    if("[I".equals(typeName)) typeName = "int[]"; else
	    if("[J".equals(typeName)) typeName = "long[]"; else
	    if("[S".equals(typeName)) typeName = "short[]"; else
	    if("[Z".equals(typeName)) typeName = "boolean[]";
            result += "    "+typeName+" "+pv.name()+";\n";
        }
        result += "}\n";
        return result;
    }

    public void setHeader(String s) {
      if(s == null) return;
      // 1. Check if there is a \programVariables section (I can
      // imagine an unlikely situation where it's actually missing),
      // if so, replace it with a "fresh" one
      int start = s.indexOf("\\programVariables");
      int end = -1;
      if(start != -1) {
         end = s.indexOf("}",start);
	 header = s.substring(0,start) + createProgramVarsSection() +
     	    s.substring(end+1);
      }else{
        // 2. No? Then put it somewhere "on top"
	start = s.indexOf("\\withOptions");
	if (start != -1) {
	  // Put it after \withOptions
	  start = s.indexOf(";",start) + 1;
	} else {
	  // Put it after \javaSource
	  start = s.indexOf("\\javaSource");
	  if(start != -1) {
	    start = s.indexOf(";",start) + 1;
	  } else {
	    Debug.fail("Your .key file is missing vital sections as far as this situation is concerned.");
	  }
	}
	 header = s.substring(0,start) + createProgramVarsSection() +
     	    s.substring(start+1);
      }
    }

    /**
     * returns the method lemma factory used to create a lemma rule for
     * this contract
     */
    protected AbstractMethodLemmaFactory getMethodLemmaFactory() {
        return new DLMethodLemmaFactory(getPre(), getPost(), sta,
                                        ruleSet, getModifies(),
                                        getSchemaModality(), name,
                                        displayName,
                                        services);
    }

    /**
     * creates a lemma rule and registers it using the appropriate
     * justification in the proof environment.
     */
    public Taclet getLemmaAndRegister() {
        if (lemma==null) {
            AbstractMethodLemmaFactory ms = getMethodLemmaFactory();
            lemma=ms.getTaclet();
            if (lemma!=null) {
                env.registerRule(lemma, new RuleJustificationBySpec(this));
            }
        }
        return lemma;
    }

    private String getName() {
        return (name == null ? "DL Hoare Triple": name)
            +((lemma==null) ? "" : " "+lemma.displayName());
    }

    /** returns a proof obligation input needed to show the corretcness of
     * this contract.
     *
     */
    public ProofOblInput getProofOblInput() {
        return new DLHoareTriplePO(fma, getModality(), header, getName(), services);
    }

    

    public String toString() {
        return "[DL Hoare Triple Contract] " + 
	    name == null ? displayName == null ? lemma == null ? "(unnamed)" : 
	    lemma.displayName() : displayName : name;
    }

    /**
     * returns the object on which this contract is, in this case it is the
     * program statement of the contract; or in case that this is a method body
     * statement the according program method is returned.
     */
    public Object getObjectOfContract() {
        Object o = unwrapBlocks(getStatement(), false);
        if (o instanceof MethodBodyStatement) {
            return ((MethodBodyStatement)o).getProgramMethod(services);
        }
        return getStatement();
    }

    public Statement getStatement() {
        return sta;
    }

    static Statement unwrapBlocks(Statement s, boolean stopAtCatchAll) {
        while (s instanceof StatementBlock
                || (s instanceof CatchAllStatement && !stopAtCatchAll)
                || s instanceof Catch) {
            s = ((StatementContainer)s).getStatementAt(0);
        }
        return s;
    }

    /** represents a proof obligation input to the prover for a DLHoareTripleContract
     *
     */
    static class DLHoareTriplePO implements ProofOblInput {

        private ProofAggregate res;
        private Term term;
        private InitConfig initConfig;
        private String name;
        private Contractable[] contractable;
        private String header; // needed to loaded the proof later

        public DLHoareTriplePO(Term t, Modality mod, String header, 
                               String name, Services services) {
            this.term = desugar(t, mod);
            this.name = name;
            this.header = header;
            Statement pe  = (Statement) t.sub(1).javaBlock().program();
            Statement unwrapped = unwrapBlocks(pe, false);
            if (unwrapped instanceof MethodBodyStatement) {
                contractable = new Contractable[]{
                        ((MethodBodyStatement)unwrapped).
                        getProgramMethod(services)};
            } else if (unwrapped instanceof Contractable){
                contractable = new Contractable[]{(Contractable)unwrapped};
            } else {
                throw new IllegalStateException("DL Hoare Triple Contract defined "
                                               +"on illegal statement "+unwrapped);
            }
        }

        private static Term desugar(Term fma, Modality mod) {
            TermFactory tf = TermFactory.DEFAULT;
            StatementBlock sta = (StatementBlock)fma.sub(1).javaBlock().program();
            Statement unwrapped = unwrapBlocks(sta,true);
            if (unwrapped instanceof Desugarable) {
                sta = (StatementBlock)((Desugarable)unwrapped).desugar();
            }
            Term result = tf.createJunctorTerm
                    (Op.IMP, fma.sub(0),
                     tf.createProgramTerm(mod,
                                          JavaBlock.createJavaBlock(sta),
                                          fma.sub(1).sub(0)));

            return result.equals(fma) ? fma : result;
        }

        /** returns the proof obligation term as result of the proof obligation
         * input. If there is still no input available because nothing has
         * been read yet null is returned.
         */
        public ProofAggregate getPO() {
            return res;
        }

        /** starts reading the input and modifies the InitConfig of this
         * object with respect to the given modification
         * strategy.
         */
        public void read(ModStrategy mod) throws ProofInputException {
            Proof p = new Proof(name(), term, header,
                                initConfig.createTacletIndex(),
                                initConfig.createBuiltInRuleIndex(),
                                initConfig.getServices(),
                                initConfig.mergedProofSettings());
            res = ProofAggregate.createProofAggregate(p, name());

        }

        public void readProblem(ModStrategy mod) throws ProofInputException {
            read(mod);
        }

        public void setExceptionHandler(KeYExceptionHandler keh){
        }

        public boolean askUserForEnvironment() {
            return false;
        }

        /** reads the Java model that belongs to this input.
         */
        public String readJava() throws ProofInputException {
            return "";
        }

        /** returns the path to the Java model.
         */
        public String getJavaPath() throws ProofInputException {
            return "";
        }

        /** set the initial configuration used to read an input. It may become
         * modified during reading depending on the modification strategy used
         * for reading.
         */
        public void setConfig(InitConfig i) {
            this.initConfig = i;
        }

        public void readSpecs(){}

        public SetOfChoice readActivatedChoices() throws ProofInputException {
            return SetAsListOfChoice.EMPTY_SET;
        }

        public SetOfChoice getActivatedChoices() throws ProofInputException {
            return initConfig.getActivatedChoices();
        }

        /** reads the include section and returns an Includes object.
         */
        public Includes readIncludes() throws ProofInputException {
            return null;
        }

        /** returns the name of the proof obligation input.
         */
        public String name() {
            return name;
        }

        /**
         * returns the object for which this proof obligation proves a contract on.
         * In this case we assume that the first program element of the second
         * subterm of the
         * proof obligation is a method body statement of which the according program
         * method is returned.
         */
        public Contractable[] getObjectOfContract() {
            return contractable;
        }

        /**
         * initialises the contract, that is, it adds the proof created by this PO input
         * to the contract if the objects of contract of both the contract object and the
         * PO input are matching. Then true is returned. Otherwise nothing is done and
         * false is returned.
         */
        public boolean initContract(Contract ct) {
            if (!(ct.getObjectOfContract() instanceof ProgramMethod)) {
                return false;
            }
            if (!((ProgramMethod)ct.getObjectOfContract()).equals(
                                    getObjectOfContract()[0])) {
                return false;
            }
            ct.addCompoundProof(getPO());
            return true;
        }

        public void startProtocol() {
            // do nothing
        }

    }


}
