// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.proof.init.*;
import java.util.*;
import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;

public class OCLMethodLemmaFactory extends AbstractMethodLemmaFactory {
    
    private ReprModelMethod repr;
    private ModifiesParserHelper modPH = null;
    private InitConfig initConfig;
    private String pre;
    private String post;
    private String suffix;
    private ProgramElement proofStatement;
    private Operator modality;
    private boolean automate;

    public OCLMethodLemmaFactory(ReprModelMethod repr, InitConfig initConfig,
				 String pre, String post, String suffix,
				 boolean automate, Operator modality) {
	this.repr = repr;
	this.initConfig = initConfig;
	this.pre = pre;
	this.post = post;
	this.suffix = suffix;
	this.automate = automate;
	this.modality = modality;
	preprocessing();
    }


    protected Term getPreCond() {
	return proofPrecondition;
    }

    protected ListOfQuantifierPrefixEntry getExtraVars() {
	return proofExtraVars;
    }

    protected ListOfQuantifierPrefixEntry getQuantifierPrefix () {
        return proofQuantifierPrefix;
    }

    protected Term getExtraPre() {
	return proofPurePrefix;
    }
    
    protected Term getPostCond() {
	return proofSuffix;
    }

    protected AbstractMethodLemmaFactory.ModifiesElement[] getModifies(){        
       return modElems;                
    }

    protected String getSuffix() {
	return suffix;
    }
    

    protected RuleSet getRuleSet() {
	return (automate) ?
        (RuleSet)getInitConfig().ruleSetNS().lookup
	    (new Name("use_method_specs"))
        : null;
    }

    protected ProgramElement getProofStatement() {
	return proofStatement;
    }

    protected Operator getModality() {
	return modality;
    }

    private InitConfig getInitConfig() {
	return initConfig;
    }


    //-------- below the stuff to create a taclet that inserts this method spec 

    private OCLPostPO postPO = null;
    private ListOfQuantifierPrefixEntry proofExtraVars = null;
    private ListOfQuantifierPrefixEntry proofQuantifierPrefix = null;
    private Term proofPrecondition;
    private Term proofPurePrefix;
    private Term proofSuffix;
    private AbstractMethodLemmaFactory.ModifiesElement[] modElems;


    public void preprocessing() {

	if (repr.getMyModifClause()==null && !repr.isQuery()) {
	    return;
	}
	
	// gets the OCL-DL translated proof which is also already transformed
	// by the method described in Bastian's minor thesis so it can be
	// used in the way his formula construction needs it
	Term proof = createOCLtoDLTransformedProof();
	
	// there is an error in the pre or postcondition and therefore no
	// taclet can be created by the given rule
	if (proof == null) {
	    return;
	}
	
	// puts the eliminated quantifiers into proofQuantVars which
	// is an ArrayOfQuantifiableVariable
	Term nonQuantifiedProof = 
	    eliminateQuantificationRetainsQuantifyingVariables(proof);
	
	// partition the now unquantified proof
	final Term proofPrefix = nonQuantifiedProof.sub(0);
	proofPrecondition = proofPrefix.sub(0);//PRE,i.e.Phi-Pre before map
	proofPurePrefix   = proofPrefix.sub(1);  //Pi1
	
	proofSuffix = nonQuantifiedProof.sub(1).sub(0); //Pi2
	
	// we have to replace Op.IMP with Op.AND in the suffix, because
	// <code>AbstractMethodLemmaFactory</code> uses existential instead of
	// universal quantification for the helper variables
	// <code>getExtraVars()</code> used to get rid of @pre
	Debug.assertTrue ( proofSuffix.op () == Op.IMP );
	proofSuffix = TermFactory.DEFAULT.createJunctorTermAndSimplify
	    ( Op.AND, proofSuffix.sub ( 0 ), proofSuffix.sub ( 1 ) );
    	
	JavaBlock proofJavaBlock = nonQuantifiedProof.sub(1).javaBlock();
	createProofStatement(proofJavaBlock);
	
	// finds all modifies elements and returns a Set of them 
	modElems = AbstractMethodLemmaFactory.TermModifiesElement
                         .create(extractModifiesElements());
    }


    private Term createOCLtoDLTransformedProof() {
	// get the OCL-DL translated Term including pre- and postconditions
	postPO = new OCLPostPO(repr.getName(), repr, pre,
			       post, Op.DIA, //this is overwritten later, so it 
                                             //does not matter which modality 
                                             //is used 
			       repr.getActXmifile(), true);
	postPO.setConfig(initConfig);
	NamespaceSet nss = initConfig.namespaces();
	Namespace oldProgVarsNS = nss.programVariables();
	nss.setProgramVariables(new Namespace());
	try{
	    postPO.read(ModStrategy.NO_VARS);
	}
	// in case there is an exception we couldn't parse the given pre- or
	// postcondition and thus can't create a rule, this gets taken care of 
	// just after the call to this method in the main part
	catch (de.uka.ilkd.key.ocl.ocldltranslator.OCLTranslationException e) {
	    System.out.println("Warning: Syntax error in OCL specification of "
			       +repr.getName());
	    return null;
	} finally {
	    nss.setProgramVariables(oldProgVarsNS);
	}
	Term dlProof = postPO.getPOTerm();

	// transform it to the form which is needed for the taclet generation 
	// according to Bastian's minor thesis
    
	// first eliminate quantifiers for that    
	final QuantifierEliminator elim = QuantifierEliminator.create ( dlProof );
        proofQuantifierPrefix = elim.getQuantifiers ();
    
	Term proof = MethodSpecTransformation.getTransformedHoare
	                                (elim.getQuantifierFreeFormula());
	return proof;
    }
    

    private Term eliminateQuantificationRetainsQuantifyingVariables(Term proof){
	Term nonQuantifiedProof = proof;
	proofExtraVars = SLListOfQuantifierPrefixEntry.EMPTY_LIST;
	while (nonQuantifiedProof.op() instanceof Quantifier){
	    proofExtraVars = proofExtraVars.append
	        ( QuantifierPrefixEntry.createFor ( nonQuantifiedProof ) );
            nonQuantifiedProof = nonQuantifiedProof.sub ( 0 );
	}
	
	return nonQuantifiedProof;
    }

    private void createProofStatement(JavaBlock proofJavaBlock) {
	StatementBlock proofStatementBlock = 
	    (StatementBlock) proofJavaBlock.program();
	proofStatement = proofStatementBlock.getChildAt(0);
    }

    private SetOfTerm extractModifiesElements() {
	ProgramVariableCollector pvc 
	    = new ProgramVariableCollector(getProofStatement());
	pvc.start();
	Namespace ns = new Namespace();
	Iterator it = pvc.result().iterator();	    
	while (it.hasNext()) {
	    ns.add((Named)it.next());
	}
	modPH = new ModifiesParserHelper(repr, initConfig);
	try {
	    modPH.parseModifiesClause(ns);
	}
	// in case there is an exception we couldn't parse the given modifies
	// clause and thus can't create a rule, this gets taken care of just
	// after the call to this method in the main part
	catch (Exception e) { 
	    System.out.println("Exception while parsing modifies: "+repr+"    "+ns);
	    e.printStackTrace();
	    return null;
	}
	return modPH.getModifiesElements(); 
    }

   
}
