package de.uka.ilkd.key.rule;

import java.util.HashMap;
import java.util.LinkedList;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.proof.VariableNameProposer;
import de.uka.ilkd.key.rule.inst.ListInstantiation;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.util.Debug;

public class NewDepOnAnonUpdates implements VariableCondition {
    
    private final SchemaVariable modifies, newFunctionListSV;

    private static final TermFactory tf = TermFactory.DEFAULT;
    
    private HashMap uninterpretedFunctions = new HashMap();

    
    public NewDepOnAnonUpdates(SchemaVariable modifies, SchemaVariable list){
        this.modifies= modifies;
        this.newFunctionListSV = list;
    }

    public boolean isSchemaVariableList(SchemaVariable sv){
        return newFunctionListSV.equals(sv);
    }
    
    
    public MatchConditions check(SchemaVariable var, SVSubstitute term,
				 MatchConditions matchCond, 
				 Services services) {

        MatchConditions result = matchCond;
        SVInstantiations inst = matchCond.getInstantiations();

        if (inst.getInstantiationEntry(modifies) == null || 
               (inst.getInstantiationEntry(newFunctionListSV) instanceof ListInstantiation)){
            return result;
        }
        
        LinkedList uninterpretedFunctionsList 
                    = getUninterpretedFunctions((ListOfObject)(inst.
                            getInstantiationEntry(modifies).getInstantiation()), services);
        
        inst   = inst.addList(newFunctionListSV, uninterpretedFunctionsList.toArray());
        result = result.setInstantiations(inst);
      
        return result;
    }

 
    private LinkedList getUninterpretedFunctions(ListOfObject assignables, Services services){
        LinkedList uninterpretedFunctionsList = new LinkedList();
        final IteratorOfObject it= assignables.iterator();
        while(it.hasNext()){
            Term loc= (Term )it.next();
           
            if (loc.op() != Op.NOTHING && loc.op() != Op.EVERYTHING){                            
                final Operator locOp = instantiateModifiesElement(loc);
                
                Term con = getUninterpretedFunction(locOp, services);
                uninterpretedFunctionsList.add(con);
            }
        }
        return uninterpretedFunctionsList;   
    }
    

    /**
     * returns an uninterpreted function (wrapped in a term) used to anonymize the
     * given location. Such an uninterpreted function depends on the logic variables
     * to be found in the given list
     * @param qva the List of quantified variables the uninterpreted function depends on
     * @param loc the Term representing the location to anonymize
     * @param services the Services
     * @return the uninterpreted function for loc as term
     */
    private Term getUninterpretedFunction(Operator loc, Services services) {       
        if (uninterpretedFunctions.containsKey(loc)) {
            return (Term)uninterpretedFunctions.get(loc);
        }
        
        final Namespace functions = services.getNamespaces().functions();
        String cname = loc.name().toString();
        cname = cname.substring(cname.indexOf(".")+1);        

//      HACK
//      the called method should be deleted as soon as this is fixed

        cname = VariableNameProposer.DEFAULT.
        getNameProposalForDependingVariable("_"+cname);
        cname = 
            cname.replace(')','_').replace('(','_').replace(':','_');
        
        
        
        final Function c = new RigidFunction(new Name(cname),
                                             getSort(loc),
                                             getArgumentSorts(loc, services));
        functions.add(c);

        final Term uninterpretedFunctionAsTerm = 
            tf.createFunctionTerm(c, createTerms(c)); 
        
        uninterpretedFunctions.put(loc, uninterpretedFunctionAsTerm);
        
        return uninterpretedFunctionAsTerm;
    }

    
    
    
    /**
     * instantiates <em>one</em> element of the modifies list. If necessary new introduced
     * logic variables and guards are added to the given lists. 
     * @param modifiesElement the Term, which is the modifies element to 
     * be instantiated
     * @param guards a List where the new created guards have to be added 
     * @param quantifiedVariables a List where the new introduced logic variables are added 
     * @param serv the Services
     * @return the instantiated modifies element
     */
    private Operator instantiateModifiesElement(Term modifiesElement) {
        if (modifiesElement.op() instanceof QuantifiedArrayOp) {
            final QuantifiedArrayOp quantifiedArrayOp = 
                ((QuantifiedArrayOp)modifiesElement.op());
            return quantifiedArrayOp.isShadowed() ?
                    ShadowArrayOp.getArrayOp(quantifiedArrayOp.arraySort()) :
                        ArrayOp.getArrayOp(quantifiedArrayOp.arraySort());
        } else {                        
            return modifiesElement.op();
        }
    }
    
    /**
     * determines the arguments sorts of the given location
     */
    private Sort[] getArgumentSorts(Operator loc, Services services){
        if (loc instanceof ProgramVariable) {
            return new Sort[0];
        } else if (loc instanceof AccessOp) {
            final Sort[] argSorts = new Sort[loc.arity()];
            final Sort integerSort = services.getTypeConverter().
            getIntegerLDT().targetSort();

            if (loc instanceof AttributeOp) {
                final KeYJavaType kjt = ((ProgramVariable)
                        ((AttributeOp)loc).attribute()).
                        getContainedInKeYJavaType();
                if (services.getJavaInfo().rec2key().getSuperArrayType() == kjt &&
                        "length".equals(((AttributeOp)loc).
                                attribute().name().toString())) {
                    argSorts[0] = services.getJavaInfo().
                        getJavaLangObjectAsSort();
                } else {
                    argSorts[0] = kjt.getSort();
                }
                
            } else if (loc instanceof ArrayOp) {
                argSorts[0] = ((ArrayOp)loc).arraySort();
                argSorts[1] = integerSort;
            }

            if (((AccessOp)loc).isShadowed()) {
                argSorts[argSorts.length - 1] = integerSort; 
            }
            return argSorts;
        } else if (loc instanceof Function) {           
            final Function func = (Function) loc;               
            final Sort[] result = new Sort[func.arity()];
            
            for (int i = 0; i<result.length; i++) {
                result[i] = func.argSort(i);
            }
            return result;
        } 
        Debug.fail("Unsupported modifies element.");
        return null;
    }
    
    /**
     * 
     */
    private Sort getSort(Operator op) {
        if (op instanceof Location)  {
            return ((Location)op).sort();
        }
        Debug.fail("Unsupported modifies element.");
        return null;
    }
    
    /**
     * builds currently a dummy terms to be used as sub terms 
     * for the given function symbol
     * @return the array of created terms
     */
    private Term[] createTerms(Function f){
        final Term[] result = new Term[f.arity()];
        
        for (int i = 0; i<f.arity(); i++) {
            result[i] = tf.createFunctionTerm
                (new RigidFunction(new Name("DUMMY"), 
                        f.argSort(i), new Sort[0]));
        }
       
        return result;
    }

    
    
    public String toString(){
        return "\\newDependingOnMod("+newFunctionListSV+","+modifies+")";
    }

    public boolean isModifiesSV(SchemaVariable var) {
        return modifies == var;
    }
    
    

}
