// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.rule.updatesimplifier.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.ConstrainedFormula;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.logic.op.IUpdateOperator;


/**
 *  The update simplifier applies updates on tems/formulas. Therefore 
 * it can be initialiazed with a set of simplification rules that 
 * shall be used.
 * 
 * Calling the simplifier from outside must only be done by using 
 * the apply methods. The simplify methods are only for recursive 
 * calls made by the update simplification rules.   
 */
public class UpdateSimplifier {
    

    /**
     * List of rules used for simplifying update terms. The first
     * applicable rule is taken. 
     */
    private IUpdateRule[] simplificationRules = new IUpdateRule[0];

    /**
     * if set to true, the update simplifier is a bit more eagerly when
     * applying updates
     */
    private boolean eager = false;
    
    
    private Constraint formulaConstraint = Constraint.BOTTOM;

    public UpdateSimplifier() {       
        this(false, false);
    }
    
    /**
     * creates an instance of the update simplifier 
     * @param deletionEnable a boolean flag indicating if unused or 
     * effectless assignment pairs should be removed
     */
    public UpdateSimplifier(boolean deletionEnabled, boolean eager) {
	this.eager = eager;
        ListOfIUpdateRule usRules = SLListOfIUpdateRule.EMPTY_LIST.    
        append(new ApplyOnAnonymousUpdate(this)).
        append(new ApplyAnonymousUpdateOnNonRigid(this)).
        append(new ApplyOnUpdate(this)).
        append(new ApplyOnLocalVariableOrStaticField(this)).
        append(new ApplyOnAttributeAccess(this)).
        append(new ApplyOnArrayAccess(this)).
        append(new ApplyOnModality(this, deletionEnabled)).
        append(new ApplyOnRigidTerm(this)).
        append(new ApplyOnRigidOperatorTerm(this)).
        append(new ApplyOnNonRigidWithExplicitDependencies(this)).
        append(new ApplyOnNonRigidTerm(this));
        
        this.setSimplificationRules(usRules);        
    }

    /**
     * Uses the given rules for update simplification. The order is
     * important as the first applicable rule is taken.
     * @param rules a ListOfUpdateRule to use for update
     * simplification
     */
    public void setSimplificationRules(ListOfIUpdateRule rules) {
	simplificationRules = rules.toArray();
    }

    
    public Term simplify(Update update, Term t) {
	Term simplifiedTerm = t;
	for (int i = 0; i<simplificationRules.length; i++) {	    
	    if (simplificationRules[i].isApplicable(update, 
	                                            simplifiedTerm)) {
		return simplificationRules[i].apply(update, simplifiedTerm);		
	    }
	}
	return simplifiedTerm;
    }

    /**
     * Derive a formula that describes in which cases the top-level operator
     * (location) described by <code>target</code> is updated by the given
     * update.
     * 
     * @param update
     *            the modification of function interpretations that is
     *            considered
     * @param target
     *            description of the location we are interested in;
     *            precondition: <code>target.op() instanceof Location</code>
     * @return formula that evaluates to <tt>true</tt> iff <code>update</code>
     *         evaluates to a semantic update that affects the location
     *         described by <code>target</code>
     */
    public Term matchingCondition (Update update, Term target) {
        for ( int i = 0; i < simplificationRules.length; i++ ) {
            if ( simplificationRules[i].isApplicable ( update, target ) )
                return simplificationRules[i].matchingCondition( update, target );
        }
        Debug.fail("Don't know how to handle " + target);
        return null; // unreachable
    }
    
    /**
     * applies the update on the given term
     */
    public Term apply(Term t) {	
        formulaConstraint = Constraint.BOTTOM;
	// simplification and application part	
	return simplify(Update.createUpdate(t), t.sub(t.arity()-1));
    }


    /**
     * returns the formula constraint of the formula to be simpliefied
     * @return the formula constraint 
     */
    public Constraint formulaConstraint() {
        return formulaConstraint;
    }
    
    /**
     * 
     */
    public ConstrainedFormula apply(ConstrainedFormula target, 
				    Constraint userConstraint) {
	formulaConstraint = target.constraint();
        return target;
    }

    /**
     * @param target
     * @return
     */
    public Term simplify(Term target) {
        if (target.op() instanceof IUpdateOperator) {            
	    target = simplify ( Update.createUpdate ( target ),
				( (IUpdateOperator)target.op () ).target ( target ) );
	    if (!eager) {
		return target;
	    }
        }

        final Term[] subs = new Term[target.arity()];
        final ArrayOfQuantifiableVariable[] vars = 
            new ArrayOfQuantifiableVariable[target.arity()];
        
        boolean changed = false;
        for (int i = 0; i<subs.length; i++) {
            subs[i] = simplify(target.sub(i));
            // only true as long as no bound variables change
            vars[i] = target.varsBoundHere(i);
            if (!changed && !(subs[i].equals(target.sub(i)))) {
                changed = true;
            }
        }
        return changed ? UpdateSimplifierTermFactory.DEFAULT.
                getBasicTermFactory().
                createTerm(target.op(), subs, vars, 
                        target.javaBlock(), null) : target ;
    }
}
