// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule.conditions;


import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.SVSubstitute;
import de.uka.ilkd.key.rule.SyntacticalReplaceVisitor;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.rule.VariableConditionAdapter;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.util.Debug;


/**
 * Condition for heuristic application: Compare the instantiations
 * of two termSV or formulaSV, using a term ordering
 */
public class OrderingCondition extends VariableConditionAdapter
 {

    /**
     * The schema variables that should be compared
     * (i.e. <code>var0</code> should be smaller than
     * <code>var1</code>)
     */
    private SchemaVariable var0, var1;

    public OrderingCondition ( SchemaVariable p_var0,
			       SchemaVariable p_var1 ) {
	var0 = p_var0;
	var1 = p_var1;

	if ( !( var0.isFormulaSV () && var1.isFormulaSV () ||
		var0.isTermSV ()    && var1.isTermSV () ) )
	    throw new IllegalArgumentException ( "Illegal schema variables" );
    }

    /**
     * @return true iff the instantiation of <code>var0</code> is
     * smaller than the i. <code>var1</code> regarding the given
     * ordering
     */
    public boolean compare ( TacletApp    p_app,
			     TermOrdering p_ordering ) {
	SVInstantiations svi = p_app.instantiations ();
	Constraint       c   = p_app.constraint     ();

	Term             t0  = (Term)svi.getInstantiation ( var0 );
	Term             t1  = (Term)svi.getInstantiation ( var1 );

	Debug.assertTrue ( t0 != null && t1 != null,
			   "Trying to compare uninstantiated schema variables" );

	if ( !c.isBottom () ) {
	    // Substitute metavariables of the two terms
	    SyntacticalReplaceVisitor srv = new SyntacticalReplaceVisitor ( c );
	    t0.execPostOrder ( srv );
	    t0 = srv.getTerm ();

	    srv = new SyntacticalReplaceVisitor ( c );
	    t1.execPostOrder ( srv );
	    t1 = srv.getTerm ();
	}

	return p_ordering.compare ( t0, t1 ) < 0;
    }

    /**
     * checks if the condition for a correct instantiation is fulfilled
     * @param var the template Variable to be instantiated
     * @param term the Term to be mapped to var
     * @param svInst the SVInstantiations that are already known to be needed 
     * @return true iff condition is fulfilled
     */
    public boolean check(SchemaVariable var, SVSubstitute term, 
			 SVInstantiations svInst, Services services) {
	// This condition does not have any meaning regarding
	// soundness
	return true;
    }

    public String toString () {
	return "\\smallerThan(" + var0 + ", " + var1 + ")";
    }

}
