// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
/*
 * Created on 22.12.2004
 */
package de.uka.ilkd.key.rule.metaconstruct;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.recoderext.ImplicitFieldAdder;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.IteratorOfObject;
import de.uka.ilkd.key.rule.ListOfObject;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.util.Debug;

/**
 * Creates a quantified update for a modifies clause. This means for all
 * elements of the modifies clause a new uninterpreted function is created
 * and assigned to the locations. It also knows how to handle specific 
 * modifies elements such as quantified array operators and the everything 
 * operator. 
 */
public class IntroduceNewAnonymousUpdateOp extends AbstractMetaOperator {

    private static final TermFactory tf = TermFactory.DEFAULT;

    private Term trueTerm = tf.createJunctorTerm(Op.TRUE);

    private int updateLVCount = 0;

    /**
     * in one modifies element, a star is always instantiated with 
     * the same expression 
     */
    private Term starInstantiation = null;
    
    private HashMap uninterpretedFunctions = new HashMap();

    public IntroduceNewAnonymousUpdateOp() {
        super(new Name("#introNewAnonUpdate"), 3);
    }

    /* (non-Javadoc)
     * @see de.uka.ilkd.key.logic.op.MetaOperator#calculate(de.uka.ilkd.key.logic.Term, de.uka.ilkd.key.rule.inst.SVInstantiations, de.uka.ilkd.key.java.Services)
     */
    public Term calculate(Term term, SVInstantiations svInst, Services services) {    	
        Object[] uninFunctions;
        ListOfTerm lt = SLListOfTerm.EMPTY_LIST;
	ListOfTerm locs = SLListOfTerm.EMPTY_LIST;
	ListOfTerm guards = SLListOfTerm.EMPTY_LIST;
	LinkedList qvl = new LinkedList();
	ModifiesSV anon = (ModifiesSV) term.sub(2).op();
	ModifiesSV modifiesSV = (ModifiesSV) term.sub(0).op();
	IteratorOfObject it = ((ListOfObject) svInst.getInstantiation(modifiesSV))
						.iterator();
	
        int i=0;
        // clear map (important!)
        uninterpretedFunctions.clear();

        uninFunctions = (((ListOfObject) svInst.getInstantiation(anon)).toArray());
        
        while(it.hasNext()){
            Term loc= (Term)it.next();
   
	    if(loc.op() == Op.EVERYTHING){
	        return tf.createAnonymousUpdateTerm(AnonymousUpdate.
						    getNewAnonymousOperator(),
						    new Term[] {term.sub(1)});
	    } else if (loc.op() != Op.NOTHING){                
                final List locSpecificGuards = new LinkedList();
                final List locSpecificQuantifiedVariables = new LinkedList();
                
                // ATTENTION: Assumption in one modifies element the star operator is
                // always instantiated with the same logic variable 
                // in *.arr[*] the second star is unfolded already in the parser
                // to [0..*.arr.length-1]
                starInstantiation = null;
                loc = instantiateModifiesElement(loc, locSpecificGuards, 
                                                 locSpecificQuantifiedVariables, services);
                
                final Term con = 
                    getAnonymizer((Function)((Term)uninFunctions[i]).op(), loc);
		
                lt = lt.prepend(con);
		locs = locs.prepend(loc);		
                guards = guards.prepend(createBigConjunction(locSpecificGuards));
		qvl.addFirst(new ArrayOfQuantifiableVariable(locSpecificQuantifiedVariables));
	        i++;
            }
	}
        
        final Term quanUpdate = 
	    tf.createQuanUpdateTerm((ArrayOfQuantifiableVariable[]) 
				    qvl.toArray(new ArrayOfQuantifiableVariable[0]),
				    guards.toArray(), locs.toArray(), 
				    lt.toArray(), term.sub(1)); 
                
	return quanUpdate;
    }
    

    /**
     * creates a big conjunction of all the terms in the list
     * @param terms the List of Term to be connected by a conjunction 
     * @return the conjunction of terms
     */
    private Term createBigConjunction(List terms) {
        Term conjunction = trueTerm;
        
        final Iterator it = terms.iterator();
        
        while (it.hasNext()) {
            conjunction = tf.createJunctorTermAndSimplify(Op.AND, conjunction, 
                                                          (Term) it.next());
        }
        
        return conjunction;
    }



    
    /**
     * instantiates <em>one</em> element of the modifies list. If necessary new introduced
     * logic variables and guards are added to the given lists. 
     * @param modifiesElement the Term, which is the modifies element to 
     * be instantiated
     * @param guards a List where the new created guards have to be added 
     * @param quantifiedVariables a List where the new introduced logic variables are added 
     * @param serv the Services
     * @return the instantiated modifies element
     */
    private Term instantiateModifiesElement(Term modifiesElement, 
            List guards, List quantifiedVariables, Services serv) {
                
        
        final Term[] subs = new Term[modifiesElement.arity()];
        
        for (int i = 0, ar = modifiesElement.arity(); i<ar; i++) {
             
            subs[i] = instantiateModifiesElement(modifiesElement.sub(i), guards, 
                                                 quantifiedVariables, serv);
                                    
            if (subs[i].op() == Op.EVERYTHING) {
               Debug.assertTrue(modifiesElement.op() instanceof AttributeOp, 
                                "Unsupported modifies element: ", modifiesElement);               
               final ProgramVariable attribute = 
                   (ProgramVariable)((AttributeOp)modifiesElement.op()).attribute();                                            
 
               if (starInstantiation != null) {
                   subs[i] = starInstantiation;
               } else {               
                   subs[i] = instantiateStarAttributeAccess(attribute, guards, 
                                                            quantifiedVariables, 
                                                            serv);
                   starInstantiation = subs[i];
               }
            }           
        }
        
        if (modifiesElement.op() instanceof QuantifiedArrayOp) {
            return instantiateQuantifiedArrayAccess((QuantifiedArrayOp)modifiesElement.op(),
                                                    subs, guards, quantifiedVariables, serv);
        } else {                        
            return tf.createTerm(modifiesElement.op(), subs, 
                                 (ArrayOfQuantifiableVariable)null, null, null);
        }
    }


    /**
     * called if a modifies element contains a quantified array access which needs
     * to be instantiated. Thereby a modifies element <tt>a[lower..upper]</tt> is
     * instantiated as <tt>a[idx]</tt> (where <tt>idx</tt> is a new logic variable) 
     * and a guard restricting the range of idx is added.  
     * 
     * The new created guards and logic variables are added to the corresponding list. 
     * @param qa the QuantifiedArrayOp to be instantiated
     * @param quantSubs the already instantiated subterms  
     * @param guards the List of terms where to add the new guads 
     * @param quantifiedVars the List of logic variables where the new introduced
     * are appended to
     * @param serv the Services
     * @return the instantiated term
     */
    private Term instantiateQuantifiedArrayAccess(QuantifiedArrayOp qa, 
            Term[] quantSubs, List guards, List quantifiedVars, Services serv) {
        
        final Term subs[] = new Term[quantSubs.length-1];
        
        // reference
        subs[0] = quantSubs[0];
               
        //      index 
        final LogicVariable lv = 
            new LogicVariable(new Name("idx_"+updateLVCount++), 
                              quantSubs[1].sort());
        quantifiedVars.add(lv);        
        
        final Term index = tf.createVariableTerm(lv);        
        subs[1] = index;
                               
        guards.add(getQuantifiedArrayAccessGuard(index, 
                                                 quantSubs[1], 
                                                 quantSubs[2], 
                                                 serv));

        // transaction number if necessary
        if (qa.isShadowed()) subs[2] = quantSubs[2];
               
        return tf.createArrayTerm(qa.isShadowed() ? 
                ShadowArrayOp.getShadowArrayOp(subs[0].sort()) : 
                    ArrayOp.getArrayOp(subs[0].sort()),                                  
                    subs);
    }

    /**
     * creates the guard for a a quantified array access, i.e. for 
     * something like <code>arr[lowerBound...upperBound]</code>     
     * @param index the Term that described the index. The created guards 
     * restricts the allowed range of the index to the allowed interval. 
     * @param lowerBound the Term representing the lower bound of the interval
     * @param upperBound the Term representing the lower bound of the interval
     * @param serv the Services
     * @return the guard for the quantified array expression, i.e. sth. similar to
     * <code>lowerBound&lt;=index AND index &gt;= upperBound </code>
     */
    private Term getQuantifiedArrayAccessGuard(Term index, 
            Term lowerBound, Term upperBound, Services serv) {
                
        final Function leq = (Function) serv.getNamespaces().
             functions().lookup(new Name("leq"));            
        
        return tf.createJunctorTermAndSimplify
        (Op.AND, 
         tf.createFunctionTerm(leq, new Term[]{lowerBound, index}),
         tf.createFunctionTerm(leq, new Term[]{index, upperBound}));
    }

    /**
     * instantiates the a star attribute access and adds the needed guard to 
     * the list of guards
     * @param attribute the ProgramVariable representing the attribute, which is accessed
     * @param guards the List of terms representing the guards which take care that only 
     * the created objects are considered; a guard restricting the instantiation of star
     * is added
     * @param quantifiedVariables the List where the new introduced logic variables are 
     * appended to 
     * @param serv the Services
     * @return an instantiation of star, i.e. a term containg a 
     * logic variable <tt>lv</tt> as top level operator. The sort of lv is the one where 
     * the given attribute is declared 
     */
    private Term instantiateStarAttributeAccess(ProgramVariable attribute, 
            List guards, List quantifiedVariables, Services serv) {
                                            
        final KeYJavaType kjt = attribute.getContainedInKeYJavaType(); 
        
        String baseName = kjt.getFullName();
        baseName = baseName.substring(baseName.lastIndexOf('.')+1).toLowerCase();        
        
        // create variable representing the star 
        final LogicVariable lv = 
            new LogicVariable(new Name(baseName + "_" + updateLVCount++), 
                              kjt.getSort());
                
        quantifiedVariables.add(lv);
        final Term starReplacement = tf.createVariableTerm(lv);
                               
        // add guard
        guards.add(getStarAttributeAccessGuard(starReplacement, serv));
        
       
        return starReplacement;
    }
    
    
    /**
     * returns the guard for an star attribute access modifies element, i.e. for
     * something like <tt>*.attribute@(classname)</tt>
     * @param starReplacement the Term used to replace the {@link Op.EVERYTHING} 
     * operator
     * @param serv the Services 
     * @return the guard <code> starReplacement.<created> = TRUE </code>
     */
    private Term getStarAttributeAccessGuard(Term starReplacement, Services serv) {
        
        final ProgramVariable created = 
            serv.getJavaInfo().getAttribute(ImplicitFieldAdder.IMPLICIT_CREATED, 
                                            serv.getJavaInfo().getJavaLangObject());
        return tf.createEqualityTerm
        (tf.createAttributeTerm(created, starReplacement), 
         serv.getTypeConverter().getBooleanLDT().getTrueTerm());                
    }

    
    /**
     * returns the anonymizer term for the given location using the given 
     * anonymizerfunction symbol
     * @param anonymizerFunction an uninterpreted Function symbol
     * @param loc the Term describing the location to be anonymized
     * @return the anonymizing term
     */
    private Term getAnonymizer(Function anonymizerFunction, Term loc) {
        Debug.assertTrue(anonymizerFunction.arity() == loc.arity(), 
                "The anonymizer function has a wrong arity.");
        final Term[] subs = new Term[anonymizerFunction.arity()];
        for (int i = 0; i<loc.arity(); i++) {
            subs[i] = loc.sub(i);
        }
        return tf.createFunctionTerm(anonymizerFunction, subs);
    }
    
    /* (non-Javadoc)
     * @see de.uka.ilkd.key.logic.op.Operator#validTopLevel(de.uka.ilkd.key.logic.Term)
     */
    public boolean validTopLevel(Term term) {        
        return term.arity()==3 && term.sub(1).sort()==Sort.FORMULA;
    }

    /* (non-Javadoc)
     * @see de.uka.ilkd.key.logic.op.Operator#sort(de.uka.ilkd.key.logic.Term[])
     */
    public Sort sort(Term[] term) {        
        return term[1].sort();
    }

    /* (non-Javadoc)
     * @see de.uka.ilkd.key.logic.op.Operator#isRigid(de.uka.ilkd.key.logic.Term)
     */
    public boolean isRigid(Term term) {
        return false;
    }
    
    /** (non-Javadoc)
     * by default meta operators do not match anything 
     * @see de.uka.ilkd.key.logic.op.Operator#match(Operator, de.uka.ilkd.key.rule.MatchConditions, de.uka.ilkd.key.java.Services)
     */
    public MatchConditions match(Operator op, MatchConditions mc,
            Services services) {
        return null;
    }

}
